from __future__ import annotations

import multiprocessing
import pathlib
import threading
from typing import TYPE_CHECKING, Callable, List, Optional, Tuple

from PyQt5.QtCore import QBuffer, QByteArray, QIODevice, QPoint, QRectF
from PyQt5.QtGui import (
    QBitmap,
    QColor,
    QFontDatabase,
    QImage,
    QPainter,
    QPixmapCache,
    QRegion,
)
from PyQt5.QtWidgets import QApplication, QGraphicsScene

from neoscore.core import env, math_helpers
from neoscore.core.color import Color
from neoscore.core.exceptions import FontRegistrationError, ImageExportError
from neoscore.core.key_event import KeyEvent
from neoscore.core.mouse_event import MouseEvent
from neoscore.core.point import Point
from neoscore.core.propagating_thread import PropagatingThread
from neoscore.core.rect import Rect, RectDef
from neoscore.core.units import Inch, Mm
from neoscore.interface.brush_interface import BrushInterface
from neoscore.interface.qt import file_paths
from neoscore.interface.qt.converters import (
    color_to_q_color,
    qt_point_to_point,
    rect_to_qt_rect_f,
)
from neoscore.interface.qt.main_window import MainWindow
from neoscore.interface.qt.viewport import Viewport
from neoscore.interface.repl import running_in_ipython_gui_repl

if TYPE_CHECKING:
    from neoscore.core.document import Document

_RENDER_IMAGE_THREAD_MAX = multiprocessing.cpu_count()
_INCHES_PER_METER: float = Inch(1) / Mm(1000)
_QT_PIXMAP_CACHE_LIMIT_KB = 200_000


class AppInterface:
    """The primary interface to the application state.

    This holds much of the global application state. An ``AppInterface`` must be created
    near the start of neoscore programs.
    """

    _QT_FONT_ERROR_CODE = -1

    def __init__(
        self,
        document: Document,
        repl_refresh_func: Callable[[float], float],
        background_brush: BrushInterface,
        auto_viewport_interaction_enabled: bool,
    ):
        self.document = document
        args = ["TestApplication", "-platform", "offscreen"] if env.HEADLESS else []
        self.app = QApplication(args)
        self.main_window = MainWindow()
        self.scene = QGraphicsScene()
        self.view: Viewport = self.main_window.graphicsView
        self.view.setScene(self.scene)
        self.background_brush = background_brush
        self.auto_viewport_interaction_enabled = auto_viewport_interaction_enabled
        self.font_database = QFontDatabase()
        self.repl_refresh_func = repl_refresh_func
        self.render_image_thread_semaphore = threading.Semaphore(
            _RENDER_IMAGE_THREAD_MAX
        )
        self._viewport_rotation = 0

    def set_refresh_func(self, refresh_func: Callable[[float], float]):
        """Set a function to run automatically on a timer in the main window."""
        self.main_window.refresh_func = refresh_func

    def set_mouse_event_handler(self, handler: Callable[[MouseEvent], None]):
        """Set a function to run on mouse events."""
        self.main_window.graphicsView.mouse_event_handler = handler

    def set_key_event_handler(self, handler: Callable[[KeyEvent], None]):
        """Set a function to run on keyboard input events."""
        self.main_window.graphicsView.key_event_handler = handler

    def show(
        self,
        min_size: Optional[Tuple[int, int]] = None,
        max_size: Optional[Tuple[int, int]] = None,
        fullscreen: bool = False,
    ):
        """Open a window showing a preview of the document.

        Args:
            min_size: An optional ``(width, height)`` minimum window size tuple.
            max_size: An optional ``(width, height)`` maximum window size tuple.
            fullscreen: Whether to show the window in fullscreen mode.
                This doesn't mix well with ``max_window_size``.
        """
        self._optimize_for_interactive_view()
        self.main_window.show(min_size, max_size, fullscreen)
        if running_in_ipython_gui_repl():
            # Do not run app.exec_() in GUI REPL mode, since IPython
            # manages the GUI thread in that case.
            if not self.main_window.refresh_func:
                self.main_window.refresh_func = self.repl_refresh_func
        else:
            self.app.exec_()

    def render_image(
        self,
        rect: Optional[RectDef],
        dest: str | pathlib.Path | bytearray,
        dpi: int,
        quality: int,
        bg_color: Color,
        autocrop: bool,
        preserve_alpha: bool,
    ) -> PropagatingThread:
        """Render the scene, or part of it, to a saved image.

        This renders on the main thread but autocrops and saves the image
        on a spawned thread which is returned to allow efficient rendering
        of many images in parallel. ``render_image`` will block if too many
        render threads are already running.

        Args:
            rect: The part of the document to render, in document coordinates.
                If ``None``, the entire scene will be rendered.
            dest: An output file path or a bytearray to save to. If a bytearray
                is given, the output format will be PNG.
            dpi: The pixels per inch of the rendered image.
            quality: The quality of the output image for compressed
                image formats. Must be either ``-1`` (default compression) or
                between ``0`` (most compressed) and ``100`` (least compressed).
            bg_color: The background color for the image.
            autocrop: Whether to crop the output image to tightly
                fit the contents of the frame. If true, the image will be
                cropped such that all 4 edges have at least one pixel not of
                ``bg_color``.
            preserve_alpha: Whether to preserve the alpha channel. If false,
                some non-transparent ``bg_color`` should be provided.

        Raises:
            ImageExportError: If Qt image export fails for unknown reasons.

        """
        dpm = AppInterface._dpi_to_dpm(dpi)
        scale = dpm / Mm(1000).base_value
        if rect:
            source_rect = rect_to_qt_rect_f(Rect.from_def(rect))
        else:
            source_rect = self.scene.sceneRect()
        pix_width = int(source_rect.width() * scale)
        pix_height = int(source_rect.height() * scale)

        if preserve_alpha:
            q_image_format = QImage.Format_ARGB32
        else:
            q_image_format = QImage.Format_RGB32

        q_image = QImage(pix_width, pix_height, q_image_format)
        q_image.setDotsPerMeterX(dpm)
        q_image.setDotsPerMeterY(dpm)
        q_color = color_to_q_color(bg_color)
        q_image.fill(q_color)

        painter = QPainter()
        painter.begin(q_image)
        painter.setRenderHint(QPainter.Antialiasing)

        target_rect = QRectF(q_image.rect())

        self.scene.render(painter, target=target_rect, source=source_rect)
        painter.end()

        def finalize():
            with self.render_image_thread_semaphore:
                final_image = (
                    AppInterface._autocrop(q_image, q_color) if autocrop else q_image
                )
                if isinstance(dest, bytearray):
                    output_array = QByteArray()
                    qbuf = QBuffer(output_array)
                    qbuf.open(QIODevice.OpenModeFlag.WriteOnly)
                    success = final_image.save(qbuf, quality=quality, format="PNG")
                    qbuf.close()
                    dest.clear()
                    dest.extend(output_array)
                else:
                    success = final_image.save(
                        file_paths.resolve_qt_path(dest), quality=quality
                    )
                if not success:
                    dest_description = (
                        "bytearray" if isinstance(dest, bytearray) else dest
                    )
                    raise ImageExportError(
                        "Unknown error occurred when exporting image to "
                        + dest_description
                    )

        thread = PropagatingThread(target=finalize)
        thread.start()
        return thread

    def destroy(self):
        """Destroy the window and all global interface-level data."""
        self.app.exit()
        self.app = None
        self.scene = None

    def register_font(self, font_file_path: str | pathlib.Path) -> List[str]:
        """Register a font file with the graphics engine.

        Args:
            font_file_path: A path to a font file. Supports TrueType and OpenType formats.

        Returns:
            A list of font families found in the font.

        Raises:
            FontRegistrationError:
                If the registration fails for any reason.
        """
        font_file_path = file_paths.resolve_qt_path(font_file_path)
        font_id = self.font_database.addApplicationFont(font_file_path)
        if font_id == AppInterface._QT_FONT_ERROR_CODE:
            raise FontRegistrationError(font_file_path)
        family_names = self.font_database.applicationFontFamilies(font_id)
        if not len(family_names):
            # I think this should be impossible, but log a warning just in case
            print(f"Warning: font at {font_file_path} provided no family names")
        return self.font_database.applicationFontFamilies(font_id)

    @property
    def background_brush(self) -> BrushInterface:
        """The brush used to paint the scene background"""
        return self._background_brush

    @background_brush.setter
    def background_brush(self, value: BrushInterface):
        self._background_brush = value
        self.scene.setBackgroundBrush(value.qt_object)

    @property
    def auto_viewport_interaction_enabled(self) -> bool:
        """Whether mouse and scrollbar viewport interaction is enabled"""
        return self._auto_viewport_interaction_enabled

    @auto_viewport_interaction_enabled.setter
    def auto_viewport_interaction_enabled(self, value: bool):
        self._auto_viewport_interaction_enabled = value
        self.view.set_auto_interaction(value)

    @property
    def viewport_center_pos(self) -> Point:
        """The interactive viewport's center position in document space."""
        # Working out the center position from the transform matrix and window
        # dimensions is pretty tricky, so hand the job to Qt.
        return qt_point_to_point(self.view.mapToScene(self.view.rect().center()))

    @viewport_center_pos.setter
    def viewport_center_pos(self, value: Point):
        self.view.centerOn(value.x.base_value, value.y.base_value)

    @property
    def viewport_scale(self) -> float:
        """The interactive viewport's scale (zoom).

        Values should be greater than 0, with 1 as the base zoom and
        larger numbers zooming in.
        """
        # Deriving the scale from the transform matrix is a headache, so compute it by
        # mapping a vector of known length.
        p1 = self.view.mapToScene(QPoint(0, 0))
        p2 = self.view.mapToScene(QPoint(1, 0))
        return 1 / math_helpers.dist((p1.x(), p1.y()), (p2.x(), p2.y()))

    @viewport_scale.setter
    def viewport_scale(self, value: float):
        transform = self.view.viewportTransform()
        relative_scale_factor = value / self.viewport_scale
        self.view.setTransform(
            transform.scale(relative_scale_factor, relative_scale_factor)
        )

    @property
    def viewport_rotation(self) -> float:
        """Set the interactive viewport's rotation angle in degrees.

        The viewport is rotated about its center.
        """
        # Track the rotation explicitly to prevent headaches and ambiguities trying to
        # derive rotation from the viewport transform matrix.
        return self._viewport_rotation

    @viewport_rotation.setter
    def viewport_rotation(self, value: float):
        transform = self.view.viewportTransform()
        if self._viewport_rotation:
            transform = transform.rotate(-self._viewport_rotation)
        self.view.setTransform(transform.rotate(value))
        self._viewport_rotation = value

    def _remove_all_loaded_fonts(self):
        """Remove all fonts registered with ``register_font()``.

        This is primarily useful for testing purposes.
        """
        success = self.font_database.removeAllApplicationFonts()
        if not success:
            raise RuntimeError("Failed to remove application fonts.")

    def clear_scene(self):
        """Clear the QT Scene. This should be called before each render."""
        self.scene.clear()

    def _optimize_for_interactive_view(self):
        QPixmapCache.setCacheLimit(_QT_PIXMAP_CACHE_LIMIT_KB)
        self.view.setViewportUpdateMode(3)  # NoViewportUpdate
        self.scene.setItemIndexMethod(-1)  # NoIndex

    @staticmethod
    def _dpi_to_dpm(dpi: int) -> int:
        """Convert a Dots Per Inch value to Dots Per Meter"""
        return int(dpi / _INCHES_PER_METER)

    @staticmethod
    def _autocrop(q_image: QImage, q_color: QColor) -> QImage:
        """Automatically crop a qt image around the pixels not of a given color.

        Returns a newly cropped image; the original is left unmodified.
        """
        mask = q_image.createMaskFromColor(q_color.rgb())
        crop_rect = QRegion(QBitmap.fromImage(mask)).boundingRect()
        return q_image.copy(crop_rect)



from typing import Optional

from neoscore.core.brush import BrushDef
from neoscore.core.layout_controllers import NewLine
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.pen import PenDef
from neoscore.core.point import Point
from neoscore.core.positioned_object import render_cached_property
from neoscore.core.units import ZERO, Unit
from neoscore.western.clef_type import ClefType, ClefTypeDef
from neoscore.western.staff import Staff
from neoscore.western.staff_object import StaffObject


class Clef(MusicText, StaffObject):

    """A graphical and logical staff clef.

    These are drawn at the initially specified position, and at the beginning of new
    lines in the parent :obj:`.Staff` until a new ``Clef`` is encountered or the end of
    the ``Staff``.

    ``Staff`` uses these to map pitches to vertical positions.
    """

    def __init__(
        self,
        pos_x: Unit,
        staff: Staff,
        clef_type: ClefTypeDef,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            pos_x: The x position in the staff
            staff: The parent staff
            clef_type: The type of clef. String names of common clefs may be
                given as a convenience; see :obj:`.ClefTypeDef`.
            font: The font used. Defaults to the staff's font.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
        """
        StaffObject.__init__(self, staff)
        # Init with placeholder y position and text; clef_type setter will update
        MusicText.__init__(self, (pos_x, ZERO), staff, "", font, brush, pen)
        self.clef_type = clef_type

    @property
    def clef_type(self) -> ClefType:
        """The type of clef, both logical and graphical."""
        return self._clef_type

    @clef_type.setter
    def clef_type(self, value: ClefTypeDef):
        self._clef_type = ClefType.from_def(value)
        if callable(self._clef_type.staff_pos):
            staff_pos = self._clef_type.staff_pos(self.staff.line_count)
        else:
            staff_pos = self._clef_type.staff_pos
        self.y = self.staff.unit(staff_pos)
        self.text = self._clef_type.glyph_name
        if callable(self.clef_type.middle_c_staff_pos):
            middle_c_staff_pos = self.clef_type.middle_c_staff_pos(
                self.staff.line_count
            )
        else:
            middle_c_staff_pos = self.clef_type.middle_c_staff_pos
        self._middle_c_staff_position = self.staff.unit(middle_c_staff_pos)

    @render_cached_property
    def breakable_length(self) -> Unit:
        """Find the length in the staff during which this clef is active.

        This is defined as the distance relative to the staff until
        the next clef is encountered. If no further clefs are found,
        this is the remaining length of the staff.
        """
        # Iterate through the (typically cached) staff clef list
        # The list is sorted by x position relative to the staff,
        # which means the first loop pass in which self_staff_x
        # has been assigned must be the next clef in the staff.
        self_staff_x = None
        for staff_x, clef in self.staff.clefs:
            if self_staff_x is not None:
                return staff_x - self_staff_x
            if clef is self:
                self_staff_x = staff_x
        # No later clefs exist; return the remaining staff length
        return self.staff.breakable_length - self_staff_x

    @property
    def middle_c_staff_position(self) -> Unit:
        """The vertical position of middle C for this clef

        0 means exactly at the top staff line.
        Positive values extend *downward* below the top staff line
        while negative values extend *upward* above the top staff line.

        This value is primarily useful in calculations of pitch staff positions
        which take a clef into account
        """
        return self._middle_c_staff_position

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        if fringe_layout.pos_x_in_staff == self.pos_x_in_staff:
            pos = Point(pos.x + fringe_layout.clef, pos.y)
        super().render_complete(pos, flowable_line)

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        if fringe_layout.pos_x_in_staff == self.pos_x_in_staff:
            pos = Point(pos.x + fringe_layout.clef, pos.y)
        super().render_complete(pos, flowable_line)

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        super().render_complete(Point(pos.x + fringe_layout.clef, pos.y), flowable_line)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        super().render_complete(Point(pos.x + fringe_layout.clef, pos.y), flowable_line)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, cast

from neoscore.core.layout_controllers import MarginController, NewLine
from neoscore.core.music_font import MusicFont
from neoscore.core.pen import Pen
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject, render_cached_property
from neoscore.core.units import ZERO, Mm, Unit, make_unit_class
from neoscore.western.abstract_staff import AbstractStaff
from neoscore.western.staff_fringe_layout import StaffFringeLayout
from neoscore.western.staff_group import StaffGroup

if TYPE_CHECKING:
    from neoscore.western.tab_clef import TabClef

_LINE_SPACE_TO_FONT_UNIT_RATIO: float = 2 / 3


class TabStaff(AbstractStaff):
    """A staff for writing guitar tablature with any number of strings.

    This class is not suitable for use with :obj:`.Chordrest`, :obj:`.Clef`,
    :obj:`.TimeSignature`, and other such classes dependent on classical staff
    semantics.

    While ``TabStaff`` has a :obj:`.MusicFont`, its unit does not necessarily correspond
    to the distance between staff lines. By default, the line spacing is wider than
    classical staves, and its ``MusicFont`` is sized to 2/3 that spacing.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        length: Unit,
        group: Optional[StaffGroup] = None,
        line_spacing: Unit = Mm(2.5),
        line_count: int = 6,
        music_font: Optional[MusicFont] = None,
        pen: Optional[Pen] = None,
    ):
        """
        Args:
            pos: The position of the top-left corner of the staff
            parent: The parent for the staff. Make this a :obj:`.Flowable`
                to allow the staff to run across line and page breaks.
            length: The horizontal width of the staff
            group: The staff group this belongs to. Set this if being used in a system
                of multiple staves.
            line_spacing: The distance between two lines in the staff.
            line_count: The number of lines in the staff.
            music_font: The font to use for :obj:`.MusicText` objects in the staff.
                Unlike in :obj:`.Staff`, this font's ``unit`` is not necessarily equivalent
                to the space between two staff (string) lines. By default, this will
                use the system-wide default music font with a unit sized to 2/3 the
                staff line spacing.
            pen: The pen used to draw the staff lines. Defaults to a line with
                thickness from the music font's engraving default.
        """
        music_font = music_font or MusicFont(
            "Bravura",
            make_unit_class(
                "TabStaffTextUnit",
                line_spacing.base_value * _LINE_SPACE_TO_FONT_UNIT_RATIO,
            ),
        )
        pen = pen or Pen(thickness=music_font.engraving_defaults["staffLineThickness"])
        super().__init__(
            pos, parent, length, group, line_spacing, line_count, music_font, pen
        )

    def string_y(self, string: int) -> Unit:
        """Return the Y position of a given string's line.

        Strings are indicated from 1 to N, where 1 is the top string
        and N is the bottom.
        """
        return self.line_spacing * (string - 1)

    @property
    def font_to_staff_space_ratio(self) -> float:
        """Conversion ratio between the font's unit and the staff line spacing."""
        return cast(float, self.unit(1) / self.line_spacing)

    @render_cached_property
    def clefs(self) -> List[Tuple[Unit, TabClef]]:
        """All the clefs in this staff, ordered by their relative x pos."""
        return self.find_ordered_descendants_with_attr("_neoscore_tab_clef_type_marker")

    def active_clef_at(self, pos_x: Unit) -> Optional[TabClef]:
        """Return the active clef at a given x position, if any."""
        return next(
            (clef for (clef_x, clef) in reversed(self.clefs) if clef_x <= pos_x),
            None,
        )

    def register_layout_controllers(self):
        """Register any flowable margin controllers needed by the staff.

        Staff subclasses must implement this.
        """
        flowable = self.flowable
        if not flowable:
            return
        staff_flowable_x = flowable.descendant_pos_x(self)
        flowable.add_margin_controller(
            MarginController(
                staff_flowable_x, self.unit(StaffGroup.RIGHT_PADDING), "neoscore_staff"
            )
        )
        for clef_x, clef in self.clefs:
            flowable_x = staff_flowable_x + clef_x
            margin_needed = clef.bounding_rect.width + self.unit(
                StaffGroup.CLEF_LEFT_PADDING
            )
            flowable.add_margin_controller(
                MarginController(flowable_x, margin_needed, "neoscore_clef")
            )

    def fringe_layout_for_isolated_staff(
        self, location: Optional[NewLine]
    ) -> StaffFringeLayout:
        """Determine the staff fringe layout of this staff in isolation.

        This is the layout needed if this staff was alone in a staff system.
        ``StaffGroup`` uses this as a starting point in its fringe layout logic.

        Staff subclasses must implement this.
        """
        if location:
            staff_pos_x = location.flowable_x - self.flowable.descendant_pos_x(self)
            if staff_pos_x < ZERO:
                # This happens on the first line of a staff positioned at x>0 relative
                # to its flowable.
                staff_pos_x = ZERO
        else:
            staff_pos_x = ZERO
        # Work right-to-left through different fringe layers
        current_x = -self.unit(StaffGroup.RIGHT_PADDING)
        clef_fringe_pos = current_x
        clef = self.active_clef_at(staff_pos_x)
        if clef:
            current_x -= clef.bounding_rect.width
            clef_fringe_pos = current_x
            current_x -= self.unit(StaffGroup.CLEF_LEFT_PADDING)
        staff_fringe_pos = current_x
        return StaffFringeLayout(
            staff_pos_x, staff_fringe_pos, clef_fringe_pos, ZERO, ZERO
        )



from enum import Enum


class PenPattern(Enum):

    """Pen stroke patterns.

    The enum int values are for internal purposes and not guaranteed by the API.
    """

    # These values align with Qt's values http://doc.qt.io/qt-5.7/qt.html#PenStyle-enum

    INVISIBLE = 0
    """No pen pattern. This is equivalent to a fully transparent color.

    Instead of using this pattern directly, you may want to use :obj:`.Pen.no_pen()`
    instead.
    """

    SOLID = 1
    """A solid line"""

    DASH = 2
    """Dashes separated by blank space"""

    DOT = 3
    """Small dots separated by blank space"""

    DASHDOT = 4
    """Alternating dashes and dots"""

    DASHDOTDOT = 5
    """A repeating pattern of one dash and two dots"""



from __future__ import annotations

from typing import Optional, Union

from neoscore.core.brush import BrushDef
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.pen import PenDef
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.western.chordrest import Chordrest


class Tremolo(MusicText):
    """A tremolo marking over a single Chordrest.

    The tremolo indicator accepts either an integer for declaring the number
    of strokes of a combining (common) tremolo, or a
    SMuFL glyphname e.g. "pendereckiTremolo".

    If a bridging tremolo is required between two chordrests, this class may not be
    suitable; instead consider manually building such tremolos using :obj:`.Beam`
    objects.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        indication: Union[int, str],
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            pos: The starting position
            parent: The parent for the starting position. If no font is given, this or
                one of its ancestors must implement :obj:`.HasMusicFont`.
            indication: The type of tremolo to draw, either a stroke count for conventional
                tremolos (1-5) or an arbitrary SMuFL glyphname.
            font: If provided, this overrides any font found in the ancestor chain.
            brush: The brush to fill shapes with.
            pen: The pen to draw outlines with.
        """

        if isinstance(indication, int):
            if indication < 1 or indication > 5:
                raise ValueError(f"Invalid stroke number: {indication}")
            glyphname = "tremolo" + str(indication)
        else:
            glyphname = indication

        MusicText.__init__(self, pos, parent, glyphname, font, brush, pen)

    @classmethod
    def for_chordrest(
        cls,
        chordrest: Chordrest,
        indication: Union[int, str],
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ) -> Tremolo:
        """Convenience constructor for a tremolo attached to a chord

        Args:
           chordrest: The chord to attach the tremolo to.
           indication: The type of tremolo to draw, either a stroke count for conventional
                tremolos (1-5) or an arbitrary SMuFL glyphname.
           font: If provided, this overrides any font found in the ancestor chain.
           brush: The brush to fill shapes with.
           pen: The pen to draw outlines with.
        """
        pos = chordrest.tremolo_attachment_point
        return Tremolo(pos, chordrest, indication, font, brush, pen)



from typing import Optional, Union

from neoscore.core.layout_controllers import NewLine
from neoscore.core.music_text import MusicText
from neoscore.core.point import Point
from neoscore.core.positioned_object import PositionedObject, render_cached_property
from neoscore.core.units import ZERO, Unit
from neoscore.western import clef_type
from neoscore.western.accidental_type import AccidentalType
from neoscore.western.key_signature_type import KeySignatureType
from neoscore.western.staff import Staff
from neoscore.western.staff_object import StaffObject


class KeySignature(PositionedObject, StaffObject):

    """A logical and graphical key signature.

    The signature will be rendered initially at the given ``pos_x``,
    and at the beginning of subsequent lines until a new
    ``KeySignature`` is encountered.

    This does not currently support natural (cancelling key signatures).
    """

    # Type sentinel used to hackily check type
    # without importing the type, risking cyclic imports.
    _neoscore_key_signature_type_marker = True

    def __init__(
        self,
        pos_x: Unit,
        staff: Staff,
        key_signature_type: Union[KeySignatureType, str],
    ):
        """
        Args:
            pos_x (Unit): The x position relative to the parent staff.
            staff (Staff): The parent staff
            key_signature_type (KeySignatureType or str): A description of the
                key signature. Any KeySignatureType may be used, or a str
                of one's name.
        """
        PositionedObject.__init__(self, Point(pos_x, ZERO), staff)
        StaffObject.__init__(self, staff)
        self._key_signature_type = (
            key_signature_type
            if isinstance(key_signature_type, KeySignatureType)
            else KeySignatureType[key_signature_type.upper()]
        )

    @property
    def key_signature_type(self) -> KeySignatureType:
        """KeySignatureType: A logical description of the key signature."""
        return self._key_signature_type

    @property
    def breakable_length(self) -> Unit:
        """Key signatures extend until another is found in the staff."""
        return self.staff.distance_to_next_of_type(self)

    @render_cached_property
    def visual_width(self) -> Unit:
        """The visual width of this key signature

        This assumes that key signatures have the same width in any clef, and that the
        accidentals used in key signatures are 1 staff unit wide.
        """
        max_x = 0
        for letter, accidental_type in self.key_signature_type.value.items():
            if accidental_type is None:
                continue
            if accidental_type == AccidentalType.SHARP:
                pos_tuple = clef_type.TREBLE.key_signature_sharp_layout[letter]
            else:
                pos_tuple = clef_type.TREBLE.key_signature_flat_layout[letter]
            max_x = max(max_x, pos_tuple[0])
        # Add max position + 1 for accidental width
        return self.staff.unit(max_x + 1)

    def _render_occurrence(
        self, pos: Point, flowable_line: Optional[NewLine], for_line_start: bool
    ):
        inside_flowable = bool(flowable_line)
        # when inside flowable this pos is absolute, otherwise relative
        base_x = pos.x
        base_y = pos.y
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        if for_line_start:
            base_x += fringe_layout.key_signature
        clef = self.staff.active_clef_at(fringe_layout.pos_x_in_staff)
        for letter, accidental_type in self.key_signature_type.value.items():
            if accidental_type is None:
                continue
            if accidental_type == AccidentalType.SHARP:
                pos_tuple = clef.clef_type.key_signature_sharp_layout[letter]
            else:
                pos_tuple = clef.clef_type.key_signature_flat_layout[letter]
            acc_pos = Point(
                base_x + self.staff.unit(pos_tuple[0]),
                base_y + self.staff.unit(pos_tuple[1]),
            )
            parent = None if inside_flowable else self
            accidental = MusicText(
                acc_pos, parent, accidental_type.value, self.staff.music_font
            )
            accidental.render()
            accidental.remove()

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        for_line_start = False
        if flowable_line:
            if flowable_line.flowable_x == flowable_x:
                for_line_start = True
        elif self.x == ZERO:
            for_line_start = True
        self._render_occurrence(pos, flowable_line, for_line_start)

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        for_line_start = fringe_layout.pos_x_in_staff == self.pos_x_in_staff
        self._render_occurrence(pos, flowable_line, for_line_start)

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        self._render_occurrence(pos, flowable_line, True)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        self._render_occurrence(pos, flowable_line, True)



from __future__ import annotations

from backports.cached_property import cached_property

from neoscore.core.glyph_info import GlyphInfo
from neoscore.core.music_char import MusicChar
from neoscore.core.music_font import MusicFont


class RawMusicChar(MusicChar):
    """A music char defined directly by a unicode character"""

    def __init__(self, font: MusicFont, codepoint: str):
        """
        Args:
            font: The character's font
            codepoint: The character's raw unicode string representation.
        """
        self._codepoint = codepoint  # noqa
        super().__init__(font, "[RAW CHAR]")

    @cached_property
    def glyph_info(self) -> GlyphInfo:
        """SMuFL metadata for this character"""
        return GlyphInfo(self.glyph_name, self._codepoint, "", None, None, None)



from typing import List, NamedTuple, Optional, Union, cast

from neoscore.core.brush import Brush, BrushDef
from neoscore.core.directions import DirectionX, DirectionY
from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.music_font import MusicFont
from neoscore.core.pen import Pen, PenDef
from neoscore.core.point import ORIGIN, Point
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO, Unit
from neoscore.western.beam import Beam
from neoscore.western.chordrest import Chordrest


class _BeamState(NamedTuple):
    """The state of a beam group at a ``Chordrest`` position"""

    flag_count: int
    break_depth: Optional[int] = None
    """Indicates a subgroup break after this position.

    This indicates the number of beams to cut a subdivision to after this beam position.
    The value must be less than ``flag_count`` and greater than 0.

    For example, this encodes four 32nd notes subdivided into 2 groups of 2 connected by
    a single beam: ``[_BeamState(3), _BeamState(3, break_depth=1), _BeamState(3), _BeamState(3)]``

    Because the beam resolver is not meter-aware, it does not perform subdivision breaks
    unless explicitly requested with this field.
    """

    hook: Optional[DirectionX] = None
    """Direction for beamlet hooks.

    If provided as an override, this only has an effect if the beam
    position requires a hook and that hook direction is
    ambiguous. This only applies in places where, within a beam
    subgroup, a position has more flags than its previous and
    following position, *and* those adjacent values are equal. For
    example, a 16th note surrounded by two 8th notes.
    """


def _resolve_beam_hooks(specs: List[_BeamState]) -> List[_BeamState]:
    """Determine which states need hooks, accounting for overrides where sensible."""
    if len(specs) < 2:
        raise ValueError("Beam groups must have at least 2 members")
    states: List[_BeamState] = []
    for i, (current_count, break_depth, current_hint_hook) in enumerate(specs):
        prev_count, prev_hint_break_depth, _ = (
            specs[i - 1] if i > 0 else (None, None, None)
        )
        next_count, next_hint_break_depth, _ = (
            specs[i + 1] if i < len(specs) - 1 else (None, None, None)
        )
        # Resolve hook
        hook = None
        if prev_count is None or (prev_hint_break_depth and next_count):
            # First item in group or subgroup
            next_count = cast(int, next_count)
            if current_count > next_count:
                hook = DirectionX.RIGHT
        elif next_count is None or (break_depth and break_depth < current_count):
            # Last item in group
            prev_count = cast(int, prev_count)
            if current_count > prev_count:
                hook = DirectionX.LEFT
        else:
            # Item in middle of group
            if current_count > prev_count and current_count > next_count:
                if prev_count < next_count:
                    hook = DirectionX.RIGHT
                elif prev_count > next_count:
                    hook = DirectionX.LEFT
                else:
                    # Surrounding positions flag counts are equal, so
                    # hook direction is ambiguous. Allow override.
                    hook = current_hint_hook or DirectionX.LEFT
        states.append(_BeamState(current_count, break_depth, hook))
    return states


class _BeamPathSpec(NamedTuple):
    """An intermediate representation of beam paths."""

    depth: int
    """The beam's 1-indexed vertical position in the beam stack.

    1 represents the outermost beam, 2 represents 1 inward, and so
    on. For example, in a beam stack connecting sixteenth notes,
    ``depth=1`` would represent the 8th note beam while ``depth=2`` would
    represent the 16th note beam. In this way, ``depth`` corresponds to
    the flag counts of represented durations.
    """

    start: int
    """Index of the starting position"""

    end: Union[int, DirectionX]
    """Index of the ending position or a hook direction.

    If this is an index, it should be greater than ``start``.
    """


def _resolve_beam_layout(states: List[_BeamState]) -> List[_BeamPathSpec]:
    """Given a list of individual note beam states, work out how to render the beams."""
    # Iterate through beams from depth 0 to end, left to right.
    path_specs = []
    max_flag_count = max((state.flag_count for state in states))
    for depth in range(1, max_flag_count + 1):
        start_idx = None
        for i, state in enumerate(states):
            if start_idx is None:
                if state.flag_count >= depth:
                    start_idx = i
                else:
                    continue
            break_ends_beam = (
                state.break_depth is not None and state.break_depth < depth
            )
            next_depth_ends_beam = (
                i == len(states) - 1 or states[i + 1].flag_count < depth
            )
            if break_ends_beam or next_depth_ends_beam:
                # Beam ends after this state
                if start_idx == i:
                    # Beam only spanned one state, treat it as a hook
                    # Sanity check, hook should always be provided when required.
                    assert state.hook
                    path_specs.append(_BeamPathSpec(depth, start_idx, state.hook))
                else:
                    path_specs.append(_BeamPathSpec(depth, start_idx, i))
                start_idx = None
    return path_specs


class _BeamGroupLine(NamedTuple):
    """The line running along the outside of a group's outermost beam."""

    start_y: Unit
    """The starting y position relative to the staff.

    When the starting x position is implied to be 0, this is line's y-intercept.
    """

    slope: float


def _resolve_beam_group_line(
    chordrests: List[Chordrest], direction: DirectionY, font: MusicFont
) -> _BeamGroupLine:
    unit = chordrests[0].staff.unit
    first = chordrests[0]
    last = chordrests[-1]
    beam_group_height = _resolve_beam_group_height(chordrests, font)
    # Determine slope from the notes furthest on the side opposite of beam
    if direction == DirectionY.DOWN:
        slope_start_ref_note = first.highest_notehead
        slope_end_ref_note = last.highest_notehead
    else:
        slope_start_ref_note = first.lowest_notehead
        slope_end_ref_note = last.lowest_notehead
    if slope_start_ref_note.y > slope_end_ref_note.y:
        delta_y = unit(-1)
    elif slope_start_ref_note.y == slope_end_ref_note.y:
        delta_y = ZERO
    else:
        delta_y = unit(1)
    delta_x = last.map_to(first).x
    slope = delta_y / delta_x
    # Now find the note closest to the beam's side
    if direction == DirectionY.DOWN:
        cr_with_closest_note = max(chordrests, key=lambda c: c.lowest_notehead.y)
        cr_x = first.map_x_to(cr_with_closest_note)
        closest_y = cr_with_closest_note.lowest_notehead.y
        nearest_beam_intersect = Point(cr_x, closest_y + unit(2.5) + beam_group_height)
    else:
        cr_with_closest_note = min(chordrests, key=lambda c: c.highest_notehead.y)
        cr_x = first.map_x_to(cr_with_closest_note)
        closest_y = cr_with_closest_note.highest_notehead.y
        nearest_beam_intersect = Point(cr_x, closest_y - unit(2.5) - beam_group_height)
    # Given a beam intersect and a slope, find the beam y at ``start``
    # y = m(x - x1) + y1, where x = 0
    start_y = (slope * (-nearest_beam_intersect.x)) + nearest_beam_intersect.y
    return _BeamGroupLine(start_y, slope)


def _beam_layer_height(font: MusicFont):
    """Determine the height of a beam and its vertical padding in a given font."""
    return (
        font.engraving_defaults["beamSpacing"]
        + font.engraving_defaults["beamThickness"]
    )


def _resolve_beam_group_height(chordrests: List[Chordrest], font: MusicFont) -> Unit:
    """Find the vertical height occupied by a beam group spanning the given Chordrests.

    This determines the maximum beam depth required and uses it to
    calculate the expected maximum height in the group.
    """
    max_depth = max((c.duration.display.flag_count for c in chordrests))
    return max_depth * _beam_layer_height(font)


def _resolve_beam_direction(chordrests: List[Chordrest]) -> DirectionY:
    """Try to determine the best direction a beam group should go in.

    The algorithm works by determining the average y position of the
    outermost notes of each chord, then if that position lies above
    the middle staff line, placing the beam below (``DOWN``), and vice
    versa
    """
    middle_staff_pos = chordrests[0].staff.center_y
    s = ZERO
    for c in chordrests:
        s += c.furthest_notehead.y if c.noteheads else middle_staff_pos
    avg = s / len(chordrests)
    if avg > middle_staff_pos:
        return DirectionY.UP
    else:
        return DirectionY.DOWN


class BeamGroup(PositionedObject, HasMusicFont):
    """A beam group spanning a collection of Chordrests.

    This analyzes the given chordrests to determine a reasonable beam layout. The
    beaming algorithm does not take into account metric subdivisions; instead it
    greedily tries to beam together as many notes as possible. Subdivisions can be
    specified by setting :obj:`.Chordrest.beam_break_depth`, which indicates a break
    after the chord to the given beam count.

    While in most situations, beamlet "hooks" (as in a dotted 8th note followed by a 16th
    note) unambiguously must point right or left, there are some cases where it is
    ambiguous. For example, a 16th note between two 8th notes could have its beamlet
    point left or right. In these situations, ``BeamGroup`` will point it left by
    default, but users can override this by setting :obj:`.Chordrest.beam_hook_dir`.

    The beam direction and slant angle are determined automatically based on the given
    notes. The direction can be overridden in ``BeamGroup``'s constructor.

    Beam layout automatically modifies spanned chordrests by snapping their stems to the
    beam line and, if this causes a stem flip, correcting the chordrest component
    layout.

    This currently has some limitations:

    * It does not support beamed rests
    * It does not respond well to mutations. If being used in interactive or animated
      situations, the group likely will need to be destroyed and recreated after any
      changes affecting its chordrests.
    """

    def __init__(
        self,
        chordrests: List[Chordrest],
        direction: Optional[DirectionY] = None,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            chordrests: The notes or rests to beam across. This must have
                at least 2 items, all of which must be of durations requiring flags.
            direction: Override for the beam direction. Otherwise, the beam direction
                is automatically chosen based on the given chordrests.
            brush: The brush to fill shapes with.
            pen: The pen to draw outlines with. To ensure perfect overlaps with stems,
                this should have the same thickness of stems, derived from the
                ``MusicFont`` engraving default ``"stemThickness``.
        """
        if len(chordrests) < 2:
            raise ValueError("BeamGroup must have at least 2 Chordrests.")
        # Determine top beam path
        chordrests.sort(key=lambda c: c.x)
        self._chordrests = chordrests
        self._beams: List[Beam] = []
        super().__init__(ORIGIN, chordrests[0])
        if font is None:
            font = HasMusicFont.find_music_font(self.parent)
        self._music_font = font
        # Load engraving defaults
        self._beam_thickness = self.music_font.engraving_defaults["beamThickness"]
        self._stem_thickness = self.music_font.engraving_defaults["stemThickness"]
        # Use same pen as stem to ensure perfectly aligned overlap
        self._brush = Brush.from_def(brush) if brush else Brush()
        self._pen = Pen.from_def(pen) if pen else Pen(thickness=self._stem_thickness)
        self._direction = direction or _resolve_beam_direction(self._chordrests)
        self._create_beams()

    def _create_beams(self):
        # Work out beam direction, slope, and offset
        beam_group_line = _resolve_beam_group_line(
            self._chordrests, self.direction, self.music_font
        )
        # Adjust stems to follow group line
        for c in self._chordrests:
            # y = m(x - x1) - y1, where x = 0
            c_relative_x = c.map_x_to(self._chordrests[0])
            y = (beam_group_line.slope * c_relative_x) + beam_group_line.start_y
            # (This direction checking approach will not work for kneed beams)
            if self.direction != c.stem.direction:
                c.stem_direction = self.direction
            adjusted_stem_end_y = c.stem.map_to(self).y + y
            c.stem.end_point.y = adjusted_stem_end_y
            c.flag.remove()
            c._flag = None
        # Now create the beams!
        layer_step = _beam_layer_height(self.music_font) * -self.direction.value
        specs = BeamGroup._resolve_chordrest_beam_layout(self._chordrests)
        base_y = -self._beam_thickness if self.direction == DirectionY.DOWN else ZERO
        for spec in specs:
            start_parent = self._chordrests[spec.start].stem.end_point
            start_y = base_y + ((spec.depth - 1) * layer_step)
            if isinstance(spec.end, int):
                end_parent = self._chordrests[spec.end].stem.end_point
                end_x = ZERO
                end_y = start_y
            else:
                end_parent = start_parent
                end_x = self._beam_thickness * 2 * spec.end.value
                end_y = start_y + (-end_x * beam_group_line.slope)
            self.beams.append(
                Beam(
                    (ZERO, start_y),
                    start_parent,
                    (end_x, end_y),
                    end_parent,
                    self.music_font,
                    self._brush,
                    self._pen,
                )
            )

    @staticmethod
    def _resolve_chordrest_beam_layout(
        chordrests: List[Chordrest],
    ) -> List[_BeamPathSpec]:
        states = _resolve_beam_hooks(
            [
                _BeamState(
                    c.duration.display.flag_count, c.beam_break_depth, c.beam_hook_dir
                )
                for c in chordrests
            ]
        )
        return _resolve_beam_layout(states)

    @property
    def direction(self) -> DirectionY:
        return self._direction

    @property
    def chordrests(self) -> List[Chordrest]:
        return self._chordrests

    @property
    def beams(self) -> List[Beam]:
        return self._beams

    @property
    def music_font(self) -> MusicFont:
        return self._music_font

    def remove(self):
        # Since the beams are actually not children of the beam group
        # (which is not ideal), we need to ensure they are removed
        # when the group is.
        for beam in self.beams:
            beam.remove()
        super().remove()



from __future__ import annotations

import math
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, List, Optional, Set, Type, cast

from backports.cached_property import cached_property

from neoscore.core import neoscore
from neoscore.core.point import ORIGIN, Point, PointDef
from neoscore.core.units import ZERO, Unit
from neoscore.interface.invisible_object_interface import InvisibleObjectInterface
from neoscore.interface.positioned_object_interface import PositionedObjectInterface

if TYPE_CHECKING:
    # Used in type annotations, imported here to avoid cyclic imports
    from neoscore.core.flowable import Flowable
    from neoscore.core.layout_controllers import NewLine


class render_cached_property(cached_property):  # noqa

    """A property annotation for fields which can be cached at render time.

    You can annotate any ``PositionedObject`` property to get this behavior, including
    on inheriting classes. ::

        class Example(PositionedObject):
            @render_cached_property
            def some_expensive_computed_property(self):
                ...

    Such properties must be immutable during rendering. Typical ``@property`` setters
    are not supported.
    """

    # Note that this class extends `cached_property`, but this is just a hack to make
    # Sphinx and other tools treat it like a property.

    def __init__(self, func):
        """
        :meta private:
        """
        self.func = func
        self.attrname = None
        self.__doc__ = func.__doc__

    def __get__(self, obj, cls):  # noqa
        if obj is None:
            return self
        result = self.func(obj)
        if not getattr(obj, "_currently_rendering", None):
            return result
        property_name = self.func.__name__
        value = obj.__dict__[property_name] = result
        obj._render_cached_properties.add(property_name)  # noqa
        return value

    def __set__(self, obj, val):
        raise AttributeError(f"can't set attribute '{self.func.__name__}'")


class PositionedObject:
    """An object positioned in the scene

    This is the base class of all objects in the neoscore scene tree.

    A single ``PositionedObject`` can have multiple graphical representations derived at
    render-time. If the object's ancestor is a :obj:`.Flowable`, it will be rendered as a
    flowable object, capable of being wrapped around lines.

    The position of this object is relative to that of its parent. Each ``PositionedObject``
    has another ``PositionedObject`` for a parent, except :obj:`.Page` objects, whose parent is
    always the global root :obj:`.Document`.

    For convenience, the parent may be initialized to ``None`` to indicate the first page of
    the document.

    To place objects directly in the scene on pages other than the first, simply set the
    parent to the desired page, accessed through the global document with
    ``neoscore.document.pages[n]``
    """

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
    ):
        """
        Args:
            pos: The position of the object relative to its parent
            parent: The parent object. Defaults to the document's first page.
        """
        self.pos = pos
        self._children: List[PositionedObject] = []
        self._parent = PositionedObject._resolve_parent(parent)
        self._set_parent_and_register_self(parent)
        self._render_cached_properties: Set[str] = set()
        self._currently_rendering = False
        self._interfaces = []
        self._interface_for_children = None
        self._scale = 1.0
        self._rotation = 0.0
        self.transform_origin = ORIGIN

    @property
    def pos(self) -> Point:
        """The position of the object relative to its parent."""
        return self._pos

    @pos.setter
    def pos(self, value: PointDef):
        self._pos = Point.from_def(value)

    @property
    def scale(self) -> float:
        """A scale factor to be applied to the rendered object.

        Outside flowable contexts, scaling is inherited by children.

        Scaling occurs relative to ``self.transform_origin``, which is by default the
        local origin.
        """
        return self._scale

    @scale.setter
    def scale(self, value: float):
        self._scale = value

    @property
    def rotation(self) -> float:
        """A rotation angle in degrees.

        Outside flowable contexts, rotation is inherited by children.

        Rotation occurs relative to ``self.transform_origin``, which is by default the
        local origin.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, value: float):
        self._rotation = value

    @property
    def transform_origin(self) -> Point:
        """The origin point for rotation and scaling transforms"""
        return self._transform_origin

    @transform_origin.setter
    def transform_origin(self, value: PointDef):
        self._transform_origin = Point.from_def(value)

    @property
    def x(self) -> Unit:
        """The x position of the object relative to its parent."""
        return self.pos.x

    @x.setter
    def x(self, value: Unit):
        self.pos = Point(value, self.y)

    @property
    def y(self) -> Unit:
        """The y position of the object relative to its parent."""
        return self.pos.y

    @y.setter
    def y(self, value: Unit):
        self.pos = Point(self.x, value)

    @property
    def breakable_length(self) -> Unit:
        """The breakable length of the object.

        This is used to determine how and where rendering cuts should be made. A length
        of zero indicates that no rendering cuts will be made.

        The default implementation of this method returns ``ZERO``. Subclasses which
        want to support flowable line breaks should override this method.
        """
        return ZERO

    @property
    def parent(self) -> PositionedObject:
        """The parent object.

        If this is set to None, it defaults to the first page of the document.
        """
        return self._parent

    @parent.setter
    def parent(self, value: Optional[PositionedObject]):
        self._parent._unregister_child(self)
        self._set_parent_and_register_self(value)

    @property
    def children(self) -> List[PositionedObject]:
        """All direct children of this object."""
        return self._children

    @children.setter
    def children(self, value: List[PositionedObject]):
        self._children = value

    @property
    def descendants(self) -> Iterator[PositionedObject]:
        """All the objects in the children subtree.

        This recursively searches all the object's children
        (and their children, etc.) and provides an iterator over them.

        The current implementation performs a simple recursive DFS over
        the tree, and has the potential to be rather slow.
        """
        for child in self.children:
            for subchild in child.descendants:
                yield subchild
            yield child

    @render_cached_property
    def flowable(self) -> Optional[Flowable]:
        """The flowable this object belongs in."""
        return cast(
            Any, self.first_ancestor_with_attr("_neoscore_flowable_type_marker")  # noqa
        )

    @property
    def interfaces(self) -> List[PositionedObjectInterface]:
        """The graphical backend binding interfaces for this object

        Interface objects are created and stored here upon calling :obj:`.render`.

        Typically each ``PositionedObject`` will have at most one interface for each
        flowable line it appears in.
        """
        return self._interfaces

    @property
    def interface_for_children(self) -> Optional[PositionedObjectInterface]:
        """The low level object interface to be used by children objects.

        Outside flowable contexts, interface classes utilize a parenting scheme much
        like core classes. The interfaces of child objects should use this field as
        their parent for proper position and transform inheritance.

        Users should rarely, if ever, have to deal with this field.
        """
        return self._interface_for_children

    def descendants_of_class_or_subclass(
        self, graphic_object_class: Type[PositionedObject]
    ) -> Iterator[PositionedObject]:
        """Yield all child descendants with a given class or its subclasses."""
        for descendant in self.descendants:
            if isinstance(descendant, graphic_object_class):
                yield descendant

    def descendants_of_exact_class(
        self, graphic_object_class: Type[PositionedObject]
    ) -> Iterator[PositionedObject]:
        """Yield all child descendants of a given class, excluding sublcasses"""
        for descendant in self.descendants:
            if type(descendant) == graphic_object_class:
                yield descendant

    def descendants_with_attribute(self, attribute: str) -> Iterator[PositionedObject]:
        """Yield all child descendants which has a given attribute.

        This is useful for searching descendants for duck-typing matches.
        """
        for descendant in self.descendants:
            if hasattr(descendant, attribute):
                yield descendant

    @property
    def ancestors(self) -> Iterator[PositionedObject]:
        """All ancestors of this object.

        Follows the chain of parents up to and including the :obj:`.Document` root.

        The order begins with ``self.parent`` and traverses upward in the document tree.
        """
        ancestor = self.parent
        while True:
            yield ancestor
            if not hasattr(ancestor, "parent"):
                # Document root found
                break
            ancestor = ancestor.parent

    def first_ancestor_with_attr(self, attr: str) -> Optional[PositionedObject]:
        """Find this object's closest ancestor with an attribute"""
        return next(
            (item for item in self.ancestors if hasattr(item, attr)),
            None,
        )

    def descendant_pos(self, descendant: PositionedObject) -> Point:
        """Find the position of a descendant relative to this object.

        Raises:
            ValueError:
                If ``descendant`` is not a descendant of this object.
        """
        pos = descendant.pos
        for parent in descendant.ancestors:
            if parent == self:
                return pos
            pos += parent.pos
        raise ValueError(f"{self} is not an ancestor of {descendant}")

    def descendant_pos_x(self, descendant: PositionedObject) -> Unit:
        """Find the x position of a descendant relative to this object.

        This is a specialized version of ``descendant_pos`` provided for optimization.

        Raises:
            ValueError:
                If ``descendant`` is not a descendant of this object.
        """
        pos_x = descendant.pos.x
        for parent in descendant.ancestors:
            if parent == self:
                return pos_x
            pos_x += parent.pos.x
        raise ValueError(f"{self} is not an ancestor of {descendant}")

    def map_to(self, dst: PositionedObject) -> Point:
        """Find an object's logical position relative to this one

        This calculates the position in *logical* space, which differs from canvas space
        in that it doesn't account for repositioning of objects inside :obj:`.Flowable`
        containers. For example, this function will return the same relative position
        for two objects in a ``Flowable`` container whether they are separated by
        a line break.
        """
        # When changing this method be sure to make the equivalent change in `map_x_to`
        # Handle easy cases
        if self == dst:
            return ORIGIN
        if self.parent == dst.parent:
            return dst.pos - self.pos
        if dst.parent == self:
            return dst.pos
        if self.parent == dst:
            return -self.pos
        # Start by collecting all ancestor using IDs because they're hashable
        self_ancestor_ids = set(id(obj) for obj in self.ancestors)
        relative_dst_pos = dst.pos
        for dst_ancestor in dst.ancestors:
            if hasattr(dst_ancestor, "parent"):
                relative_dst_pos += dst_ancestor.pos
            if id(dst_ancestor) in self_ancestor_ids:
                # Now find relative_self_pos and return relative_dst_pos - relative_self_pos
                relative_self_pos = self.pos
                for self_ancestor in self.ancestors:
                    if hasattr(self_ancestor, "parent"):
                        relative_self_pos += self_ancestor.pos
                    if self_ancestor == dst_ancestor:
                        return relative_dst_pos - relative_self_pos
                # Since we've already determined there is a common
                # ancestor, this should never happen
                assert False, "Unreachable"
        raise ValueError(f"{self} and {dst} have no common ancestor")

    def map_x_to(self, dst: PositionedObject) -> Unit:
        """Like :obj:`.map_to`, but only return the X distance from to ``dst``."""

        # This implementation is copied from `map_to` and tweaked to only handle the X
        # axis. This is a very common operation, so this optimization is useful.

        # Handle easy cases
        if self == dst:
            return ZERO
        if self.parent == dst.parent:
            return dst.x - self.x
        if dst.parent == self:
            return dst.x
        if self.parent == dst:
            return -self.x
        # Start by collecting all ancestor using IDs because they're hashable
        self_ancestor_ids = set(id(obj) for obj in self.ancestors)
        relative_dst_x = dst.x
        for dst_ancestor in dst.ancestors:
            if hasattr(dst_ancestor, "parent"):
                relative_dst_x += dst_ancestor.x
            if id(dst_ancestor) in self_ancestor_ids:
                # Now find relative_self_x and return relative_dst_x - relative_self_x
                relative_self_x = self.x
                for self_ancestor in self.ancestors:
                    if hasattr(self_ancestor, "parent"):
                        relative_self_x += self_ancestor.x
                    if self_ancestor == dst_ancestor:
                        return relative_dst_x - relative_self_x
                # Since we've already determined there is a common
                # ancestor, this should never happen
                assert False, "Unreachable"
        raise ValueError(f"{self} and {dst} have no common ancestor")

    def distance_to(self, obj: PositionedObject, offset: Point = ORIGIN) -> Unit:
        """Find the distance to a given object, with an optional extra offset.

        Like :obj:`.map_to`, this works in *logical* coordinates.
        """
        relative_dst_pos = self.map_to(obj) + offset
        distance = Unit(
            math.sqrt(
                (relative_dst_pos.x.base_value**2)
                + (relative_dst_pos.y.base_value**2)
            )
        )
        return type(self.x)(distance)

    def canvas_pos(self) -> Point:
        """Find the document-space position of this object.

        For objects in :obj:`.Flowable`\ s, this should only be accessed at render time,
        when flowable layouts are available.
        """
        pos = ORIGIN
        current = self
        while hasattr(current, "parent"):
            pos += current.pos
            current = current.parent
            if hasattr(current, "map_to_canvas"):
                # Parent appears to be a flowable,
                # so let it decide where the point goes.
                return cast(Any, current).map_to_canvas(pos)
        return pos

    def remove(self):
        """Remove this object from the document tree."""
        if self.parent:
            self.parent.children.remove(self)

    def pre_render_hook(self):
        """Run code once just before document rendering begins.

        Implementations *must* call the superclass function as well.
        """
        self._currently_rendering = True

    def post_render_hook(self):
        """Run code once after document rendering completes.

        Implementations *must* call the superclass function as well.
        """
        for cached_property in self._render_cached_properties:
            del self.__dict__[cached_property]
        self._render_cached_properties.clear()
        self._currently_rendering = False

    def render(self):
        """Render the object and all its children.

        This and other render methods should generally not be called directly.
        """
        if self.flowable is not None:
            self.render_in_flowable()
        else:
            self._interface_for_children = InvisibleObjectInterface(
                self.pos,
                # Hack because root document obj lacks this property
                getattr(self.parent, "interface_for_children", None),
                self.scale,
                self.rotation,
                self.transform_origin,
            )
            self._interface_for_children.render()
            self.render_complete(self.pos)
        for child in self.children:
            child.render()

    def render_in_flowable(self):
        """Render the object to the scene, dispatching partial rendering calls
        when needed if an object flows across a break in the flowable.

        This and other render methods should generally not be called directly.
        """
        # Calculate position within flowable
        assert self.flowable is not None
        pos_in_flowable = self.flowable.descendant_pos(self)
        first_line_i = self.flowable.last_break_index_at(pos_in_flowable.x)
        first_line = self.flowable.lines[first_line_i]
        first_line_length = (
            first_line.flowable_x + first_line.length - pos_in_flowable.x
        )
        remaining_x = self.breakable_length - first_line_length
        if remaining_x <= ZERO:
            self.render_complete(self.canvas_pos(), first_line, pos_in_flowable.x)
            return

        # Render before break
        if first_line_length < Unit(1):
            # If a break-spanning object starts very close to its first line end,
            # skip that line.
            first_line_i += 1
            first_line = self.flowable.lines[first_line_i]
            first_line_length = (
                first_line.flowable_x + first_line.length - pos_in_flowable.x
            )
            remaining_x = self.breakable_length - first_line_length
        line_pos = first_line.canvas_pos()
        render_start_pos = Point(
            line_pos.x + (pos_in_flowable.x - first_line.flowable_x),
            line_pos.y + pos_in_flowable.y,
        )
        self.render_before_break(render_start_pos, first_line, pos_in_flowable.x)

        # Iterate through remaining length
        for current_line_i in range(first_line_i + 1, len(self.flowable.lines)):
            current_line = self.flowable.lines[current_line_i]
            line_pos = current_line.canvas_pos()
            render_start_pos = Point(line_pos.x, line_pos.y + pos_in_flowable.y)
            local_object_x = self.breakable_length - remaining_x
            if remaining_x > current_line.length:
                # Render spanning continuation
                self.render_spanning_continuation(
                    render_start_pos, current_line, local_object_x
                )
                remaining_x -= current_line.length
            else:
                # Render end
                self.render_after_break(render_start_pos, current_line, local_object_x)
                break

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        """Render the entire object.

        This is used to render all objects outside flowables, as well as those inside
        flowables when they fit completely in one line of the flowable.

        By default, this is a no-op. Subclasses with rendered appearances should
        override this.

        This method behaves differently inside and outside of flowables. Whether this
        object is inside a flowable can be determined by whether a ``flowable_line`` is
        given. When inside a flowable, the given position is in global document
        coordinates, and created interfaces (or higher level classes) must not be
        assigned a parent. When not inside a flowable, the given position is relative to
        ``self.parent`` and created interfaces (or higher level classes) must be
        assigned a parent. In this case, created interfaces should use
        ``self.parent.interface_for_children`` as their parent.

        This and other render methods should generally not be called directly.

        Args:
            pos: The rendering position. If outside a flowable, this is relative to
                the parent. Otherwise, it is in document coordinates.
            flowable_line: If in a ``Flowable``, the line in which this object appears
            flowable_x: If in a ``Flowable``, the flowable x position of this render

        """

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        """Render the beginning of the object up to a stopping point.

        For use in flowable containers when rendering an object that crosses a line or
        page break. This function should render the beginning portion of the object up
        to the break.

        By default, this is a no-op. Subclasses with rendered appearances should
        override this.

        Created interfaces and higher level objects should not be assigned a parent.

        This and other render methods should generally not be called directly.

        Args:
            pos: The rendering position in document space for drawing.
            flowable_line: The line in which this object appears
            flowable_x: The flowable x position of this render

        """

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        """
        Render the continuation of an object after a break and before another.

        For use in flowable containers when rendering an object that crosses
        two breaks. This function should render the portion of the object
        surrounded by breaks on either side.

        By default, this is a no-op. Subclasses with rendered appearances should
        override this.

        Created interfaces and higher level objects should not be assigned a parent.

        This and other render methods should generally not be called directly.

        Args:
            pos: The rendering position in document space for drawing.
            flowable_line: The line in which this object appears
            object_x: The local object x position of the line's start.
        """

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        """Render the continuation of an object after a break.

        For use in flowable containers when rendering an object that crosses a line or
        page break. This function should render the ending portion of an object after a
        break.

        By default, this is a no-op. Subclasses with rendered appearances should
        override this.

        Created interfaces and higher level objects should not be assigned a parent.

        This and other render methods should generally not be called directly.

        Args:
            pos: The rendering position in document space for drawing.
            flowable_line: The line in which this object appears
            object_x: The local object x position of the line's start.
        """

    @staticmethod
    def _resolve_parent(value: Optional[PositionedObject]) -> PositionedObject:
        if value is None:
            return neoscore.document.pages[0]
        return value

    def _set_parent_and_register_self(self, value: Optional[PositionedObject]):
        if value is None:
            value = neoscore.document.pages[0]
        self._parent = value
        if hasattr(self._parent, "_register_child"):
            self._parent._register_child(self)

    def _register_child(self, child: PositionedObject):
        """Add an object to ``self.children``."""
        self.children.append(child)

    def _unregister_child(self, child: PositionedObject):
        """Remove an object from ``self.children``."""
        self.children.remove(child)



from dataclasses import dataclass
from typing import Optional

from neoscore.core.units import Unit
from neoscore.interface.font_interface import FontInterface
from neoscore.interface.positioned_object_interface import PositionedObjectInterface
from neoscore.interface.qt.converters import point_to_qt_point_f
from neoscore.interface.qt.q_rich_text_item import QRichTextItem


@dataclass(frozen=True)
class RichTextInterface(PositionedObjectInterface):

    """An interface for graphical text objects."""

    html_text: str

    font: FontInterface

    width: Optional[Unit] = None

    def render(self):
        """Render the line to the scene."""
        self._register_qt_object(self._create_qt_object())

    def _create_qt_object(self) -> QRichTextItem:
        """Create and return this interface's underlying Qt object"""
        qt_object = QRichTextItem()
        qt_object.setHtml(self.html_text)
        qt_object.document().setDocumentMargin(0)
        qt_object.setPos(point_to_qt_point_f(self.pos))
        qt_object.setTextWidth(self.width.base_value if self.width is not None else -1)
        qt_object.setFont(self.font.qt_object)
        qt_object.setTransformOriginPoint(point_to_qt_point_f(self.transform_origin))
        if self.scale != 1:
            qt_object.setScale(self.scale)
        if self.rotation != 0:
            qt_object.setRotation(self.rotation)
        return qt_object



from dataclasses import dataclass

from PyQt5.QtWidgets import QGraphicsSimpleTextItem

from neoscore.core.point import ORIGIN
from neoscore.interface.positioned_object_interface import PositionedObjectInterface
from neoscore.interface.qt.converters import point_to_qt_point_f


@dataclass(frozen=True)
class InvisibleObjectInterface(PositionedObjectInterface):
    """A stub interface for use as a virtual parent in scenes."""

    def render(self):
        qt_object = QGraphicsSimpleTextItem()
        qt_object.setPos(point_to_qt_point_f(self.pos))
        if self.transform_origin != ORIGIN:
            qt_object.setTransformOriginPoint(
                point_to_qt_point_f(self.transform_origin)
            )
        if self.scale != 1:
            qt_object.setScale(self.scale)
        if self.rotation != 0:
            qt_object.setRotation(self.rotation)
        if self.transform_origin != ORIGIN:
            qt_object.setTransformOriginPoint(
                point_to_qt_point_f(self.transform_origin)
            )
        self._register_qt_object(qt_object)



"""Enums describing directions"""

from __future__ import annotations

from enum import Enum

from neoscore.core.units import Unit


class DirectionY(Enum):
    """A vertical direction"""

    UP = -1
    DOWN = 1

    def flip(self) -> DirectionY:
        """Return the opposite direction as this"""
        if self == DirectionY.UP:
            return DirectionY.DOWN
        else:
            return DirectionY.UP

    @classmethod
    def from_sign(cls, value: float | Unit) -> DirectionY:
        """Get a ``DirectionY`` from the sign of a number or ``Unit`` value."""
        if isinstance(value, Unit):
            sign = -1 if value.base_value < 0 else 1
        else:
            sign = -1 if value < 0 else 1
        if sign == -1:
            return DirectionY.UP
        else:
            return DirectionY.DOWN


class DirectionX(Enum):
    """A horizontal direction"""

    LEFT = -1
    RIGHT = 1

    def flip(self) -> DirectionX:
        """Return the opposite direction as this"""
        if self == DirectionX.LEFT:
            return DirectionX.RIGHT
        else:
            return DirectionX.LEFT

    @classmethod
    def from_sign(cls, value: float | Unit) -> DirectionX:
        """Get a ``DirectionX`` from the sign of a number or ``Unit`` value."""
        if isinstance(value, Unit):
            sign = -1 if value.base_value < 0 else 1
        else:
            sign = -1 if value < 0 else 1
        if sign == -1:
            return DirectionX.LEFT
        else:
            return DirectionX.RIGHT



from __future__ import annotations

import copy
from typing import Dict, Optional, Type, Union

from neoscore.core import neoscore, smufl
from neoscore.core.exceptions import (
    MusicFontGlyphNotFoundError,
    MusicFontMetadataNotFoundError,
)
from neoscore.core.font import Font
from neoscore.core.glyph_info import GlyphInfo
from neoscore.core.point import Point
from neoscore.core.rect import Rect
from neoscore.core.units import Mm, Unit, convert_all_to_unit, make_unit_class


class MusicFont(Font):

    """A SMuFL compliant music font"""

    def __init__(self, family_name: str, size: Union[Unit, Type[Unit]]):
        """
        Args:
            family_name: The font name
            size: The font size, given either as the distance between two staff lines,
                or a unit type where ``unit(1)`` is that distance.
        """
        if isinstance(size, Unit):
            self._unit = make_unit_class("StaffUnit", size.base_value)
        else:
            self._unit = size
        try:
            self.metadata = neoscore.registered_music_fonts[family_name]
        except KeyError:
            raise MusicFontMetadataNotFoundError
        self._engraving_defaults = copy.deepcopy(self.metadata["engravingDefaults"])
        # 1 SMuFL em is the height of a 5-line staff. See:
        # w3c.github.io/smufl/latest/specification/scoring-metrics-glyph-registration.html
        self._em_size = self.unit(4)
        self._glyph_info_cache = {}
        # engraving_defaults is small, so eagerly converting it to self.unit is ok
        convert_all_to_unit(self._engraving_defaults, self.unit)
        super().__init__(family_name, self._em_size, 1, False)

    def __str__(self):
        unit_val_as_mm = Mm(self.unit(1)).display_value
        return f"MusicFont('{self.family_name}', <unit(1) = Mm({unit_val_as_mm})>)"

    @property
    def unit(self) -> Type[Unit]:
        """A unit type where ``unit(1)`` is a standard staff space in the font."""
        return self._unit

    @property
    def engraving_defaults(self) -> Dict:
        """The SMuFL engraving defaults for this font.

        See `SMuFL's description of this data here
        <https://w3c.github.io/smufl/latest/specification/engravingdefaults.html>`_.
        """
        return self._engraving_defaults

    def modified(  # noqa
        self, family_name: Optional[str] = None, unit: Optional[Type[Unit]] = None
    ) -> MusicFont:
        return MusicFont(
            family_name if family_name is not None else self.family_name,
            unit if unit is not None else self.unit,
        )

    def glyph_info(
        self, glyph_name: str, alternate_number: Optional[int] = None
    ) -> GlyphInfo:
        """Look up metadata on a glyph in this font.

        Args:
            glyph_name: The canonical name of the glyph, or its main version
                if using an alternate number.
            alternate_number: A glyph alternate number.

        Raises:
            MusicFontGlyphNotFoundError: If the requested glyph
                could not be found in the font.
        """

        key = (glyph_name, alternate_number)
        cached_result = self._glyph_info_cache.get(key, None)
        if cached_result:
            return cached_result
        try:
            computed_result = self._glyph_info(glyph_name, alternate_number)
        except ValueError:
            raise MusicFontGlyphNotFoundError(glyph_name, alternate_number)
        self._glyph_info_cache[key] = computed_result
        return computed_result

    def _glyph_info(
        self, glyph_name: str, alternate_number: Optional[int] = None
    ) -> GlyphInfo:
        # if an alt glyph get name
        if alternate_number:
            glyph_name = self._check_alternate_names(glyph_name, alternate_number)

        # check if glyphname exists then get details from smufl
        check_name = smufl.glyph_names.get(glyph_name)
        if check_name:
            codepoint = check_name["codepoint"]
            description = check_name["description"]
        else:
            #  check is it ligature or optional glyph and get info
            (codepoint, description) = self._check_optional_glyphs(glyph_name)

        # if we have made it this far and populated
        # info with all other valid details
        advance_width = self.unit(0)
        glyphAdvanceWidths = self.metadata.get("glyphAdvanceWidths")
        if glyphAdvanceWidths:
            raw_advance_width = glyphAdvanceWidths.get(glyph_name)
            if raw_advance_width:
                advance_width = self.unit(raw_advance_width)

        bounding_rect = copy.deepcopy(self.metadata["glyphBBoxes"].get(glyph_name))
        if bounding_rect:
            convert_all_to_unit(bounding_rect, self.unit)
            bounding_rect = self._convert_bbox_to_rect(bounding_rect)

        # get optional anchor metadata if available
        anchors = self._load_glyph_anchors(glyph_name)

        return GlyphInfo(
            glyph_name, codepoint, description, bounding_rect, advance_width, anchors
        )

    @staticmethod
    def _convert_bbox_to_rect(b_box_dict: dict) -> Rect:
        """Converst the SMuFL bounding box info
        into a Rect class format"""

        # get SMuFL bbox coords
        ne_x = b_box_dict["bBoxNE"][0]
        ne_y = b_box_dict["bBoxNE"][1]
        sw_x = b_box_dict["bBoxSW"][0]
        sw_y = b_box_dict["bBoxSW"][1]

        # calculate neoscore Rect coords
        x = sw_x
        y = ne_y * -1
        width = ne_x - sw_x
        height = ne_y - sw_y

        return Rect(x=x, y=y, width=width, height=height)

    def _check_alternate_names(self, glyph_name: str, alternate_number: int) -> str:
        """Check to see if the alternate glyph exists,
        if it does it then returns that glyph name.

        Args:
            glyph_name: The canonical name of the glyph, or its main version
                if using an alternate number.
            alternate_number: A glyph alternate number
        """

        # check if glyphname has alternates
        alternate_glyphs = self.metadata["glyphsWithAlternates"].get(glyph_name)

        # Alternate not found in the font
        if not alternate_glyphs:
            raise ValueError

        else:
            # check if valid alt number
            alt_count = len(alternate_glyphs["alternates"])

            # if the alternate_number is in range
            if alt_count >= alternate_number:
                new_glyph_name = alternate_glyphs["alternates"][alternate_number - 1][
                    "name"
                ]
            else:
                # Alternate number out of range
                raise ValueError

        return new_glyph_name

    def _check_optional_glyphs(self, glyph_name: str) -> Tuple[str, str]:
        """Check to see if the called glyph exists  as an optional, if it does it then
         populate the GlyphInfo dataclass with basic info.

        Args: info: a partially populated GlyphInfo dataclass
            glyph_name: The canonical name of the glyph, or its main version
                if using an alternate number.
        """

        # else check the optional glyph list
        optional_glyph_field = self.metadata["optionalGlyphs"].get(glyph_name)

        if optional_glyph_field:
            codepoint = optional_glyph_field["codepoint"]

            # some don't have descriptions
            description = optional_glyph_field.get("description")

        # else glyphname is not registered with SMuFL
        else:
            raise ValueError

        return codepoint, description

    def _load_glyph_anchors(self, glyph_name: str) -> Optional[Dict[str, Point]]:
        """Load any glyph anchors and convert coordinates to neoscore points."""
        anchors = self.metadata["glyphsWithAnchors"].get(glyph_name)
        if anchors is None:
            return None
        anchors = copy.deepcopy(anchors)
        for key, value in anchors.items():
            # SMuFL coords have opposite Y axis as neoscore, so flip
            # when wrapping in Point and Unit.
            anchors[key] = Point(self.unit(value[0]), self.unit(-value[1]))
        return anchors



from __future__ import annotations

from typing import Optional

from neoscore.core.exceptions import DynamicStringError
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject


class Dynamic(MusicText):

    """A textual dynamic marking"""

    # Map from letters to SMuFL canonical names
    _dynamic_letter_map = {
        "p": "dynamicPiano",
        "m": "dynamicMezzo",
        "f": "dynamicForte",
        "r": "dynamicRinforzando",
        "s": "dynamicSforzando",
        "z": "dynamicZ",
        "n": "dynamicNiente",
    }

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        text: str,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            pos: Position relative to ``parent``
            parent: If no font is given, this or one of its ancestors must
                implement :obj:`.HasMusicFont`.
            text: A valid dynamic indicator string consisting
                of the letters: 'p, m, f, r, s, z, n'
            font: If provided, this overrides any font found in the ancestor chain.
        """
        parsed_text = self._parse_dynamic_string(text)
        MusicText.__init__(self, pos, parent, parsed_text, font)

    @classmethod
    def ppp(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create a 'ppp' dynamic."""
        return cls(pos, parent, "ppp", font)

    @classmethod
    def pp(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create a 'pp' dynamic."""
        return cls(pos, parent, "pp", font)

    @classmethod
    def p(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create a 'p' dynamic."""
        return cls(pos, parent, "p", font)

    @classmethod
    def mp(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create an 'mp' dynamic."""
        return cls(pos, parent, "mp", font)

    @classmethod
    def mf(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create an 'mf' dynamic."""
        return cls(pos, parent, "mf", font)

    @classmethod
    def f(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create a 'f' dynamic."""
        return cls(pos, parent, "f", font)

    @classmethod
    def ff(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create a 'ff' dynamic."""
        return cls(pos, parent, "ff", font)

    @classmethod
    def fff(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create a 'fff' dynamic."""
        return cls(pos, parent, "fff", font)

    @classmethod
    def sfz(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create an 'sfz' dynamic."""
        return cls(pos, parent, "sfz", font)

    @classmethod
    def fp(
        cls, pos: PointDef, parent: PositionedObject, font: Optional[MusicFont] = None
    ) -> Dynamic:
        """Create an 'fp' dynamic."""
        return cls(pos, parent, "fp", font)

    @classmethod
    def _parse_dynamic_string(cls, string: str) -> List[str]:
        """Parse a dynamics string into a list of SMuFL canonical names"""
        music_chars = []
        for char in string:
            try:
                music_chars.append(cls._dynamic_letter_map[char])
            except KeyError:
                raise DynamicStringError(string, char)
        return music_chars



from __future__ import annotations

from typing import Optional

from neoscore.core.brush import BrushDef
from neoscore.core.music_font import MusicFont
from neoscore.core.music_path import MusicPath
from neoscore.core.pen import PenDef
from neoscore.core.point import Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO


class Beam(MusicPath):

    """A rhythmic beam connecting groups of notes.

    This is a single beam - for multiple layers of beams (e.g. 2 for
    16th notes), multiple of these must be stacked on top of each
    other. See :obj:`.BeamGroup` for a reasonable implementation of this.

    While this is a path, it requires a music font from which to
    derive its appearance.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        end_pos: PointDef,
        end_parent: PositionedObject,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            pos: The starting (left) position of the beam
            parent: The parent for the starting position. If no font is given,
                this or one of its ancestors must implement :obj:`.HasMusicFont`.
            end_pos: The ending (right) position of the beam
            end_parent: The parent for the ending position.
                Must be a staff or in one.
            font: If provided, this overrides any font found in the ancestor chain.
            brush: The brush to fill shapes with.
            pen: The pen to draw outlines with.
        """
        MusicPath.__init__(self, pos, parent, font, brush, pen)
        self.beam_thickness = self.music_font.engraving_defaults["beamThickness"]
        # Draw beam
        end_pos = Point.from_def(end_pos)
        self.line_to(end_pos.x, end_pos.y, end_parent)
        self.line_to(end_pos.x, end_pos.y + self.beam_thickness, end_parent)
        self.line_to(ZERO, self.beam_thickness, self)
        self.close_subpath()



from enum import Enum


class BrushPattern(Enum):

    """Brush fill patterns.

    The enum int values are for internal purposes and not guaranteed by the API.
    """

    # Values align with Qt's https://doc.qt.io/qt-5/qt.html#BrushStyle-enum

    INVISIBLE = 0
    """No brush pattern. This is equivalent to a fully transparent color.

    Instead of using this pattern directly, consider using :obj:`.Brush.no_brush()`
    instead.
    """

    SOLID = 1
    """Uniform solid color."""

    DENSE_1 = 2
    """Extremely dense brush pattern."""

    DENSE_2 = 3
    """Very dense brush pattern."""

    DENSE_3 = 4
    """Somewhat dense brush pattern."""

    DENSE_4 = 5
    """Half dense brush pattern."""

    DENSE_5 = 6
    """Somewhat sparse brush pattern."""

    DENSE_6 = 7
    """Very sparse brush pattern."""

    DENSE_7 = 8
    """Extremely sparse brush pattern."""

    HORIZONTAL_LINES = 9
    """Horizontal lines."""

    VERTICAL_LINES = 10
    """Vertical lines."""

    CROSSING_VERTICAL_HORIZONTAL_LINES = 11
    """Crossing horizontal and vertical lines."""

    ASCENDING_DIAGONAL_LINES = 12
    """Diagonal lines ascending from left to right."""

    DESCENDING_DIAGONAL_LINES = 13
    """Diagonal lines descending from left to right."""

    CROSSING_DIAGONAL_LINES = 14
    """Crossing diagonal lines."""



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional

from typing_extensions import TypeAlias

from neoscore.core.directions import DirectionX
from neoscore.core.page import Page

if TYPE_CHECKING:
    from neoscore.core.document import Document


PageOverlayFunc: TypeAlias = Callable[[Page], None]
"""A function to run on every page after creation.

The function takes one argument, the newly generated page. Functions will typically
create objects with this page as their parent.

See :obj:`.simple_header_footer` for a ready-made simple overlay function.
"""


class PageSupplier:
    """A supplier and generator-on-demand of document :obj:`.Page` objects.

    This acts like a list of ``Page`` objects which generates them as needed.
    Externally, it can be used mostly as a list. If an index is requested for which no
    page yet exists, that page will be generated, as well as any missing pages between
    the previous last page and the one requested. Consequently, keep in mind that
    innocent looking operations such as ``page_suppler[100000]`` are actually expensive
    operations, as they implicitly generate thousands of Page objects.

    The contents of the ``PageSupplier`` should be treated as immutable. Attempts to
    modify the pages it contains will likely result in unexpected behavior.

    This is an internal class meant to be created by the global :obj:`.Document` for its
    ``pages`` property.
    """

    def __init__(
        self, document: Document, overlay_func: Optional[PageOverlayFunc] = None
    ):
        """
        Args:
            document: The global document using this object.
            overlay_func: A function to call with every page when generated.
                This can be used to create headers and footers.
        """
        self._document = document
        self._page_list: List[Page] = []
        self.overlay_func = overlay_func

    def __getitem__(self, index):
        if index >= len(self._page_list):
            for new_index in range(len(self._page_list), index + 1):
                page_side = DirectionX.LEFT if new_index % 2 else DirectionX.RIGHT
                new_page = Page(
                    self.document.page_origin(new_index),
                    self.document,
                    new_index,
                    page_side,
                    self.document.paper,
                )
                self._page_list.append(new_page)
                if self.overlay_func:
                    self.overlay_func(new_page)
        return self._page_list[index]

    def __iter__(self):
        return self._page_list.__iter__()

    def __len__(self):
        return len(self._page_list)

    @property
    def document(self) -> Document:
        return self._document

    @property
    def overlay_func(self) -> Optional[PageOverlayFunc]:
        """A function to call on every page generation.

        This function is called with every generated page at the time of generation. If
        the value is changed it will only affect pages generated after the change.
        """
        return self._overlay_func

    @overlay_func.setter
    def overlay_func(self, value: Optional[PageOverlayFunc]):
        self._overlay_func = value



from typing import List, Optional

from neoscore.core.brush import BrushDef
from neoscore.core.music_font import MusicFont
from neoscore.core.pen import PenDef
from neoscore.core.units import Unit
from neoscore.western.tab_staff import TabStaff
from neoscore.western.tab_string_text import TabStringText


class TabNumber(TabStringText):
    """A number placed in a tab staff typically used to indicate frets or fingers"""

    def __init__(
        self,
        pos_x: Unit,
        staff: TabStaff,
        string: int,
        number: int,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        hide_background=True,
    ):
        """
        Args:
            pos_x: The x position relative to the parent
            staff: The parent staff
            string: The 1-indexed string number this should appear on
            number: The number to display. Must be a non-zero integer.
            font: The font to use. Defaults to the staff's font.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
            hide_background: Whether to paint over the background behind the text.
                Particularly useful for preventing overlaps with staff lines.
        """
        TabStringText.__init__(
            self,
            pos_x,
            staff,
            string,
            TabNumber._number_to_digit_glyph_names(number),
            font,
            brush,
            pen,
            hide_background,
        )

    @staticmethod
    def _number_to_digit_glyph_names(number: int) -> List[str]:
        return [f"fingering{digit}" for digit in str(number)]



from __future__ import annotations

from typing import Optional, cast

from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.point import Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.western.duration import Duration, DurationDef
from neoscore.western.duration_display import DurationDisplay


class Rest(MusicText):

    """A simple Rest glyph whose appearance is determined by a duration"""

    _glyphnames = {
        1024: "rest1024th",
        512: "rest512th",
        256: "rest256th",
        128: "rest128th",
        64: "rest64th",
        32: "rest32nd",
        16: "rest16th",
        8: "rest8th",
        4: "restQuarter",
        2: "restHalf",
        1: "restWhole",
    }

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        duration: DurationDef,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            pos: The position of the rest from its SMuFL anchor point
            parent: The parent of the rest. If no font is provided,
                this parent or one of its ancestors must implement :obj:`.HasStaffUnit`.
            duration: Determines the particular rest glyph used.
            font: If provided, this overrides any font inherited from an ancestor.
        """
        pos = Point.from_def(pos)
        self.duration = Duration.from_def(duration)
        duration_display = cast(DurationDisplay, self.duration.display)
        MusicText.__init__(
            self,
            pos,
            parent,
            [self._glyphnames[duration_display.base_duration]],
            font,
        )

    @property
    def duration(self) -> Duration:
        """The time duration of this Rest"""
        return self._duration

    @duration.setter
    def duration(self, value: DurationDef):
        value = Duration.from_def(value)
        if value.display is None:
            raise ValueError(f"{value} cannot be represented as a single note")
        self._duration = value



"""SMuFL spec metadata loaded into module constants."""
import json
import pathlib

# Load metadata into dictionaries

_SMUFL_DIR = pathlib.Path(__file__).parent / ".." / "resources" / "smufl"

classes: dict
"""The raw SMuFL ``classes.json`` metadata.

See https://w3c.github.io/smufl/latest/specification/classes.html

:meta hide-value:
"""

glyph_names: dict
"""The raw SMuFL ``glyphnames.json`` metadata.

See https://w3c.github.io/smufl/latest/specification/glyphnames.html

:meta hide-value:
"""

ranges: dict
"""The raw SMuFL ``ranges.json`` metadata.

See https://w3c.github.io/smufl/latest/specification/ranges.html

:meta hide-value:
"""

with open(_SMUFL_DIR / "classes.json", "r") as classes_file:
    classes = json.load(classes_file)

with open(_SMUFL_DIR / "glyphnames.json", "r") as glyphnames_file:
    glyph_names = json.load(glyphnames_file)

with open(_SMUFL_DIR / "ranges.json", "r") as ranges_file:
    ranges = json.load(ranges_file)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from neoscore.core.units import ZERO, Inch, Mm, Unit


@dataclass(frozen=True)
class Paper:

    """A specification for a paper geometry used to lay out pages"""

    width: Unit
    height: Unit
    margin_top: Unit = field(default_factory=lambda: ZERO)
    margin_right: Unit = field(default_factory=lambda: ZERO)
    margin_bottom: Unit = field(default_factory=lambda: ZERO)
    margin_left: Unit = field(default_factory=lambda: ZERO)
    gutter: Unit = field(default_factory=lambda: ZERO)
    live_width: Unit = field(init=False)
    live_height: Unit = field(init=False)

    def __post_init__(self):
        # This hack is needed to assign derived fields on frozen dataclasses
        # See https://stackoverflow.com/a/54119384/5615927
        super().__setattr__(
            "live_width",
            self.width - self.gutter - self.margin_left - self.margin_right,
        )
        super().__setattr__(
            "live_height", self.height - self.margin_bottom - self.margin_top
        )

    def make_rotation(self) -> Paper:
        """Create a 90-degree clockwise rotation of this paper.

        The ``gutter`` field is left unchanged.
        """
        return Paper(
            self.height,
            self.width,
            self.margin_left,
            self.margin_top,
            self.margin_right,
            self.margin_bottom,
            self.gutter,
        )

    def modified(
        self,
        width: Optional[Unit] = None,
        height: Optional[Unit] = None,
        margin_top: Optional[Unit] = None,
        margin_right: Optional[Unit] = None,
        margin_bottom: Optional[Unit] = None,
        margin_left: Optional[Unit] = None,
        gutter: Optional[Unit] = None,
    ) -> Paper:
        """Derive a new ``Paper`` from this one with any given changed attributes."""
        return Paper(
            width if width is not None else self.width,
            height if height is not None else self.height,
            margin_top if margin_top is not None else self.margin_top,
            margin_right if margin_right is not None else self.margin_right,
            margin_bottom if margin_bottom is not None else self.margin_bottom,
            margin_left if margin_left is not None else self.margin_left,
            gutter if gutter is not None else self.gutter,
        )


# Templates for common paper types are declared below

A4 = Paper(Mm(210), Mm(297), Mm(20), Mm(20), Mm(20), Mm(20), ZERO)
"""Template for A4-sized portrait paper"""

LETTER = Paper(Inch(8.5), Inch(11), Inch(1), Inch(1), Inch(1), Inch(1), ZERO)
"""Template for letter-sized portrait paper"""



from __future__ import annotations

from typing import Optional, cast

from neoscore.core.directions import DirectionY
from neoscore.core.exceptions import NoFlagNeededError
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.western.duration import Duration, DurationDef
from neoscore.western.duration_display import DurationDisplay


class Flag(MusicText):

    """A simple flag glyph determined by a duration and direction"""

    _up_glyphnames = {
        8: "flag1024thUp",
        7: "flag512thUp",
        6: "flag256thUp",
        5: "flag128thUp",
        4: "flag64thUp",
        3: "flag32ndUp",
        2: "flag16thUp",
        1: "flag8thUp",
    }
    _down_glyphnames = {
        8: "flag1024thDown",
        7: "flag512thDown",
        6: "flag256thDown",
        5: "flag128thDown",
        4: "flag64thDown",
        3: "flag32ndDown",
        2: "flag16thDown",
        1: "flag8thDown",
    }

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        duration: Duration,
        direction: DirectionY,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            pos: The position of this flag. When ``parent`` is a stem end point
                this should typically be :obj:`.ORIGIN`.
            parent: If no font is given, this or one of its ancestors must
                implement :obj:`.HasMusicFont`.
            duration: The duration corresponding to the flag. This controls the flag
                glyph rendered.
            direction: The direction of the flag
            font: If provided, this overrides any font found in the ancestor chain.
        """
        self.duration = duration
        self.direction = direction
        duration_display = cast(DurationDisplay, self.duration.display)
        if duration_display.flag_count == 0:
            raise NoFlagNeededError(self.duration)
        if self.direction == DirectionY.DOWN:
            glyph_name = self._down_glyphnames[duration_display.flag_count]
        else:
            glyph_name = self._up_glyphnames[duration_display.flag_count]
        MusicText.__init__(self, pos, parent, [glyph_name], font)

    @property
    def duration(self) -> Duration:
        return self._duration

    @duration.setter
    def duration(self, value: DurationDef):
        value = Duration.from_def(value)
        if value.display is None:
            raise ValueError(f"{value} cannot be represented as a single note")
        self._duration = value

    @property
    def direction(self) -> DirectionY:
        return self._direction

    @direction.setter
    def direction(self, value: DirectionY):
        self._direction = value

    @classmethod
    def vertical_offset_needed(cls, duration: Duration) -> int:
        """Find the space needed in a stem using a flag of a given duration

        The value is given in pseudo-staff-units.
        """
        if duration.display is None:
            return 0
        elif duration.display.flag_count:
            return 1
        else:
            return 0



"""Application-wide environment variables.

Some global behavior can be switched by environment variable toggles. For example
running any neoscore program with the environment variable ``NEOSCORE_DEBUG`` to some
truthy value will enable some debugging information like Qt bounding box outlines.
"""

import os


def _resolve_bool_env_variable(var):
    value = os.environ.get(var)
    return not (value is None or value == "0" or value.lower() == "false")


HEADLESS = _resolve_bool_env_variable("NEOSCORE_HEADLESS")
"""Whether to run the Qt application in headless mode.

Set by the environment variable ``NEOSCORE_HEADLESS``.
"""


DEBUG = _resolve_bool_env_variable("NEOSCORE_DEBUG")
"""Whether debug mode is enabled.

Set by the environment variable ``NEOSCORE_DEBUG``.
"""



from PyQt5.QtCore import QPointF
from PyQt5.QtGui import QBrush, QColor, QPainter, QPen
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsTextItem

from neoscore.core import env
from neoscore.core.point import ORIGIN
from neoscore.interface.qt.converters import point_to_qt_point_f


class QRichTextItem(QGraphicsTextItem):
    """A lighted specialized version of ``QGraphicsTextItem``

    * Enables painter caching
    * Allows debugging bounding rect drawing
    * Enables anti-aliasing
    """

    def __init__(
        self,
        transform_origin: QPointF = ORIGIN,
    ):
        """
        Args:
            transform_origin: The origin point for rotation and scaling transforms
        """
        super().__init__()
        self.setCacheMode(QGraphicsItem.CacheMode.DeviceCoordinateCache)
        self.setTransformOriginPoint(point_to_qt_point_f(transform_origin))

    def paint(self, painter: QPainter, *args, **kwargs):
        if env.DEBUG:
            bounding_rect = self.boundingRect()
            painter.setBrush(QBrush())
            painter.setPen(QPen(QColor("#ff0000"), 0))
            painter.drawRect(bounding_rect)
        # Enable text anti-aliasing - for some reason the global render hint isn't
        # applied here.
        painter.setRenderHint(0x02)
        super().paint(painter, *args, **kwargs)



from typing import Optional

from PyQt5.QtCore import QPointF, QRectF
from PyQt5.QtGui import QBrush, QColor, QPainter, QPainterPath, QPen
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsPathItem

from neoscore.core import env


class QClippingPath(QGraphicsPathItem):

    """A QGraphicsPathItem extension supporting horizontal path clipping.

    Works like a ``QGraphicsPathItem`` except that it renders a
    horizontal slice of the path. Rather than rendering the entire
    path, renders the region starting at a given ``clip_start_x`` and
    extending for a given ``clip_width``. This rendered region is
    shifted leftward, so it appears at the path's root position. This
    is useful for splitting a path into horizontal chunks and
    rendering them in different positions, for instance when drawing a
    staff which appears on multiple lines.

    ``clip_start_x`` and ``clip_width`` should not take into account
    scaling. For example if a rendered region of 50 points is required
    on a path with a scale of 2, ``clip_width=50`` should be passed.

    While the Qt superclass is mutable, this is intended to be treated
    immutably. Mutations after instantation will result unexpected
    behavior. Object mutations at higher abstraction levels should
    result in new Qt objects created.

    Internally, the clipping implementation is rather subtle in how it
    integrates with Qt's coordinate and painter systems. The item's
    bounding rect is adjusted to match the requested clip region. At
    render time, the painter translates its coordinate system leftward
    by the (internally scale-adjusted) ``clip_start_x``. The painter's
    clip rect is then derived from the item's bounding rect, but
    shifted rightward to cancel out the painter's translation. These
    actions are all automatically scaled as necessary, since the scale
    is applied to the QClippingPath, not the painter.

    Note that clipping behavior does not play well with rotated items,
    and no API guarantees are currently given about it.
    """

    def __init__(
        self,
        qt_path: QPainterPath,
        clip_start_x: float = 0,
        clip_width: Optional[float] = None,
        scale: float = 1,
        rotation: float = 0,
        background_brush: Optional[QBrush] = None,
        defer_geometry_calculation: bool = False,
        transform_origin: Optional[QPointF] = None,
    ):
        """
        Args:
            qt_path: The path for the item. This value should
                be the same as in ``QGraphicsPathItem.__init__()``
            clip_start_x: The local starting position for the path clipping region.
                This should not adjust for scaling, as that is performed
                automatically. Use ``0`` to render from the start.
            clip_width: The width of the path clipping region. This should not adjust
                for scaling, as that is performed automatically. Use ``None`` to render
                to the end
            scale: A scaling factor on the object's coordinate system.
            rotation: Rotation about the path's origin given in degrees. Rotated path
                clipping is currently not supported.
            background_brush: If given, this will be used to paint over the path's
                bounding rect behind the path.
            defer_geometry_calculation: If true, this constructor will not automatically
                calculate the path's bounding and clipping geometry. When this is set,
                you *must* call ``update_geometry`` when the geometry is finalized.
                This is useful when post-init modifications immediately alter the geometry,
                preventing a redundant calculation.
            transform_origin: The origin point for rotation and scaling transforms
        """
        super().__init__(qt_path)
        self.clip_start_x = clip_start_x / scale
        self.clip_width = None if clip_width is None else clip_width / scale
        self.setCacheMode(QGraphicsItem.CacheMode.DeviceCoordinateCache)
        if transform_origin:
            self.setTransformOriginPoint(transform_origin)
        self.setRotation(rotation)
        super().setScale(scale)
        self.background_brush = background_brush
        self.bounding_rect = None
        self.clip_rect = None
        if not defer_geometry_calculation:
            self.update_geometry()

    def boundingRect(self):
        # Seems like this is in logical space (pre-scaling)
        return self.bounding_rect

    def paint(self, painter: QPainter, *args, **kwargs):
        """Paint with automatic clipping.

        This is overridden from ``QGraphicsPathItem.paint()``
        """
        if self.clip_start_x != 0:
            painter.translate(-self.clip_start_x, 0)
        if self.clip_rect is not None:
            painter.setClipRect(self.clip_rect)
        if env.DEBUG or self.background_brush:
            bounding_rect = self.bounding_rect
            if self.clip_start_x != 0:
                # Since painter is translated, cancel that out when
                # drawing the bounding rect
                bounding_rect = bounding_rect.translated(self.clip_start_x, 0)
        if self.background_brush:
            painter.setBrush(self.background_brush)
            painter.setPen(QPen(0))
            painter.drawRect(bounding_rect)
        if env.DEBUG:
            painter.setBrush(QBrush())
            painter.setPen(QPen(QColor("#ff0000"), 0))
            painter.drawRect(bounding_rect)

        super().paint(painter, *args, **kwargs)

    def update_geometry(self):
        """Recalculate the object's bounding and clipping rects.

        This *must* be called after any changes affecting these rects, including pen
        thickness changes.
        """
        self.prepareGeometryChange()
        path_bounding_rect = self.path().boundingRect()
        padding = self.pen().widthF() / 2
        self.bounding_rect = QClippingPath.calculate_bounding_rect(
            path_bounding_rect,
            self.clip_start_x,
            self.clip_width,
            padding,
        )
        # Clip rect is used by painter, which translates by -clip_start_x,
        # so we need to cancel that out here
        self.clip_rect = self.bounding_rect.translated(self.clip_start_x, 0)

    @staticmethod
    def calculate_bounding_rect(
        bounding_rect: QRectF,
        clip_start_x: float,
        clip_width: Optional[float],
        padding: float,
    ) -> QRectF:
        """Create a QRectF giving the bounding rect for the path.

        Args:
            bounding_rect: The full shape's bounding rectangle
            clip_start_x: The local starting position for the
                clipping region. Use ``None`` to render from the start.
            clip_width: The width of the clipping region.
                Use ``None`` to render to the end
            padding: Extra area padding to be added to all non-clipped sides
                of the rect.
        """
        # We used to do this in a more DRY way, but it was very difficult to reason
        # about which factors applied to which cases, so to make things much easier to
        # reason about we simply split each of the 4 cases apart and handle them
        # separately.
        if not clip_start_x and not clip_width:
            # Full bounding rect
            # (Either not in a flowable, or fits completely in line)
            return QRectF(
                bounding_rect.x() - padding,
                bounding_rect.y() - padding,
                bounding_rect.width() + (padding * 2),
                bounding_rect.height() + (padding * 2),
            )
        elif clip_width and (not clip_start_x):
            # Starting from beginning, using a clip width
            # (Incomplete first line in flowable)
            return QRectF(
                bounding_rect.x() - padding,
                bounding_rect.y() - padding,
                # Do not pad right edge
                (clip_width - bounding_rect.x()) + padding,
                bounding_rect.height() + (padding * 2),
            )
        elif clip_width and clip_start_x:
            # Starting from middle of path, using a clip width
            # (After first line in a flowable, but not at the end yet)
            return QRectF(
                # Do not pad left or right edge
                0.0,
                bounding_rect.y() - padding,
                clip_width,
                bounding_rect.height() + (padding * 2),
            )
        elif (not clip_width) and clip_start_x:
            # Starting from middle of path, extends to the end
            # (Last line in flowable after a continuation)
            return QRectF(
                # Do not pad left edge
                0.0,
                bounding_rect.y() - padding,
                (bounding_rect.width() - clip_start_x + bounding_rect.x()) + padding,
                bounding_rect.height() + (padding * 2),
            )
        else:
            raise RuntimeError("Unreachable")



"""Built-in page overlay helpers"""

from typing import Optional

from neoscore.core import neoscore
from neoscore.core.directions import DirectionX
from neoscore.core.font import Font
from neoscore.core.page import Page
from neoscore.core.page_supplier import PageOverlayFunc
from neoscore.core.text import Text


def simple_header_footer(
    outside_top_text: Optional[str] = None,
    centered_top_text: Optional[str] = None,
    outside_bottom_text: Optional[str] = None,
    centered_bottom_text: Optional[str] = None,
    font: Optional[Font] = None,
) -> PageOverlayFunc:
    """Create an overlay function for simple headers and footers.

    Args:
        outside_top_text: Text to display on the top outside page corner.
        centered_top_text: Text to display centered in the top margin
        outside_bottom_text: Text to display on the bottom outside page corner.
        centered_bottom_text: Text to display centered in the bottom margin
        font: Font to use for all text. If none is provided, this will default
            to a slightly smaller version of the default font.

    The outside corner positions vary based on whether the page is left or right
    sided. The text items are positioned half along the margins. Center text is
    automatically center-aligned, and outside text is aligned to the outside.

    All text items can include a magic template string ``"%page"``, which will be
    automatically replaced by the overlaid page's number. For instance,
    ``simple_header_footer(centered_bottom_text="Page #%page)`` will generate
    footers with the text "Page #1", "Page #2", etc.

    This generates and reutrns a ``PageOverlayFunc`` which should be passed to the
    document's page generator with:

    >>> neoscore.document.pages.overlay_func = simple_header_footer() # doctest: +SKIP

    This only affects pages created after set, so typically this should be done right
    after ``neoscore.setup()``.

    """

    def generated_overlay_func(page: Page):
        n = page.index + 1
        # These different resolved names are needed to prevent closure scoping issues
        top_corner_text = (
            _format_template(outside_top_text, n) if outside_top_text else None
        )
        top_center_text = (
            _format_template(centered_top_text, n) if centered_top_text else None
        )
        bottom_corner_text = (
            _format_template(outside_bottom_text, n) if outside_bottom_text else None
        )
        bottom_center_text = (
            _format_template(centered_bottom_text, n) if centered_bottom_text else None
        )
        text_font = font or neoscore.default_font.modified(
            size=neoscore.default_font.size * 0.9
        )
        if page.page_side == DirectionX.LEFT:
            outside_x = page.left_margin_x - (page.full_margin_left / 2)
        else:
            outside_x = page.right_margin_x + (page.full_margin_right / 2)
        center_x = page.center_x
        top_y = page.top_margin_y - (page.paper.margin_top / 2)
        bottom_y = page.bottom_margin_y + (page.paper.margin_bottom / 2)
        if top_corner_text:
            t = Text((outside_x, top_y), page, top_corner_text, text_font)
            if page.page_side == DirectionX.RIGHT:
                t.x -= t.bounding_rect.width
        if top_center_text:
            t = Text((center_x, top_y), page, top_center_text, text_font)
            t.x -= t.bounding_rect.width / 2
        if bottom_corner_text:
            t = Text((outside_x, bottom_y), page, bottom_corner_text, text_font)
            if page.page_side == DirectionX.RIGHT:
                t.x -= t.bounding_rect.width
        if bottom_center_text:
            t = Text((center_x, bottom_y), page, bottom_center_text, text_font)
            t.x -= t.bounding_rect.width / 2

    return generated_overlay_func


def _format_template(template: str, page_number: int) -> str:
    return template.replace("%page", str(page_number))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['neoscore',
 'neoscore.core',
 'neoscore.interface',
 'neoscore.interface.qt',
 'neoscore.western']

package_data = \
{'': ['*'],
 'neoscore': ['resources/*',
              'resources/fonts/bravura/*',
              'resources/fonts/lora/*',
              'resources/smufl/*']}

install_requires = \
['PyQt5>=5.15.6,<6.0.0',
 'backports.cached-property==1.0.2',
 'img2pdf==0.4.4',
 'sortedcontainers==2.4.0',
 'typing_extensions>=4,<5']

setup_kwargs = {
    'name': 'neoscore',
    'version': '0.1.13',
    'description': 'A graphical musical notation library',
    'long_description': '# neoscore\n\n## *notation without bars*\n\n![A score with colored blocks and squiggly lines](https://raw.githubusercontent.com/DigiScore/neoscore/main/doc/static/img/promo_image.png)\n*[Example source](https://github.com/DigiScore/neoscore/blob/main/examples/promo_image.py)*\n\nNeoscore is a Python library for creating scores without limits. While other notation software assumes scores follow a narrow set of rules, neoscore treats scores as shapes and text with as few assumptions as possible. In neoscore, staves and noteheads are just one way of writing. Its programmatic nature makes it especially useful for generative scoremaking, and it even supports experimental animation and live-coding!\n\n## Quick Start\n\nNeoscore requires a minimum [Python version of **3.7**](https://www.python.org/downloads/). If you don\'t have it you\'ll need to first install it, then [set up a virtual environment with it](https://realpython.com/python-virtual-environments-a-primer/). In that environment you can then install neoscore with pip  using `pip install neoscore`, after which you should be able to run this example:\n\n```python\nfrom neoscore.common import *\nneoscore.setup()\nText(ORIGIN, None, "Hello, neoscore!")\nneoscore.show()\n```\n\nIf you have installation problems, please see [our troubleshooting guide](https://neoscore.org/community/support.html).\n\n## [Documentation](https://neoscore.org)\n\nVisit [neoscore.org](https://neoscore.org) for thorough documentation and dozens of examples. You can find more [elaborate examples in this repository here](https://github.com/DigiScore/neoscore/blob/main/examples).\n\n------------------\n\n### Credits\n\nNeoscore was principally developed by [Andrew Yoon](https://andrewyoon.art), originally begun in 2016 at [The Recurse Center](https://www.recurse.com/) then revived and released in 2022 with substantial support from [Craig Vear](https://www.dmu.ac.uk/about-dmu/academic-staff/art-design-humanities/craig-vear/craig-vear.aspx) and the [DigiScore](https://digiscore.dmu.ac.uk/) research project. Additional support was provided by [Prashanth Thattai Ravikumar](https://github.com/prashanthtr). Many of neoscore’s design choices were informed by the excellent open source notation projects [Abjad](https://github.com/Abjad/abjad), [Lilypond](https://lilypond.org/index.html), and [MuseScore](https://github.com/musescore/MuseScore/).\n\nThe Digital Score project (DigiScore) is funded by the European Research Council (ERC) under the European Union’s Horizon 2020 research and innovation programme (Grant agreement No. ERC-2020-COG - 101002086).\n\n![Logo for the European Research Council](https://raw.githubusercontent.com/DigiScore/neoscore/main/doc/static/img/erc_eu_logo.webp)\n\nGet your name added here by getting involved!\n',
    'author': 'Andrew Yoon',
    'author_email': 'andrew@nothing-to-say.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://neoscore.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Optional

from neoscore.core import neoscore
from neoscore.core.brush import BrushDef
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicStringDef, MusicText
from neoscore.core.pen import PenDef
from neoscore.core.text_alignment import AlignmentX, AlignmentY
from neoscore.core.units import Unit
from neoscore.western.tab_staff import TabStaff


class TabStringText(MusicText):
    """SMuFL text centered on a tab staff line.

    This also automatically adds a background brush which hides the
    tab staff line behind the text.
    """

    def __init__(
        self,
        pos_x: Unit,
        staff: TabStaff,
        string: int,
        text: MusicStringDef,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        hide_background: bool = True,
        breakable: bool = True,
        alignment_x: AlignmentX = AlignmentX.CENTER,
        alignment_y: AlignmentY = AlignmentY.CENTER,
    ):
        """
        Args:
            pos_x: The x position relative to the parent. The text will be
                centered around this position.
            staff: The parent staff
            string: The 1-indexed string number this should appear on
            text: The text to display. Can be given as a SMuFL glyph name,
                or other shorthand forms. See :obj:`.MusicStringDef`.
            font: The font to use. Defaults to the staff's font.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
            hide_background: Whether to paint over the background behind the text.
                Particularly useful for preventing overlaps with staff lines.
            breakable: Whether this object should break across lines in
                Flowable containers.
            alignment_x: The text's horizontal alignment relative to ``pos``.
                Note that the default value here is center-alignment, unlike
                most other text classes.
            alignment_y: The text's vertical alignment relative to ``pos``.
                Note that the default value here is center-alignment, unlike
                most other text classes.
        """
        background_brush = neoscore.background_brush if hide_background else None
        pos_y = staff.string_y(string)
        MusicText.__init__(
            self,
            (pos_x, pos_y),
            staff,
            text,
            font,
            brush,
            pen,
            1,
            0,
            background_brush,
            breakable,
            alignment_x,
            alignment_y,
        )



from dataclasses import dataclass
from enum import Enum, auto


class KeyEventType(Enum):
    """Enum describing keyboard input event types.

    The enum integer values are arbitrary and may change in the future.
    """

    PRESS = auto()
    """A key was pressed"""

    AUTO_REPEAT = auto()
    """A key was automatically repeated from being held"""

    RELEASE = auto()
    """A key was released"""


@dataclass(frozen=True)
class KeyEvent:
    """A keyboard input event.

    For printable keys, the easiest way to read a key press is by the ``text``
    field. For non-printing keys like arrow keys, you can check the pressed key with the
    ``code`` field.
    """

    event_type: KeyEventType
    """The type of the event"""

    code: int
    """The platform-independent key code for the pressed key.
    
    This key code corresponds to `Qt's platform-independent key map
    <https://doc.qt.io/qt-5/qt.html#Key-enum>`_. For convenience, a few common key codes
    are provided here:

    +---------------+-------------+
    | Key           | Code (hex)  |
    +===============+=============+
    | <Left Arrow>  | 0x01000012  |
    +---------------+-------------+
    | <Up Arrow>    | 0x01000013  |
    +---------------+-------------+
    | <Right Arrow> | 0x01000014  |
    +---------------+-------------+
    | <Down Arrow>  | 0x01000015  |
    +---------------+-------------+
    | <Space>       | 0x20        |
    +---------------+-------------+
    | <Enter>       | 0x01000004  |
    +---------------+-------------+
    | <Escape>      | 0x01000000  |
    +---------------+-------------+

    Note that this does not differentiate letter capitalization; check the active
    modifier keys or the ``text`` field to determine capitalization.
    """

    modifiers: int
    """A bitmask of key modifiers active during this event.

    The modifier flags in this bitmask correspond to
    `Qt's modifier map <https://doc.qt.io/qt-5/qt.html#KeyboardModifier-enum>`_.
    """

    text: str
    """The unicode text generated by this key.

    This may be empty for keystrokes that do not emit text, such as modifiers.
    """



from threading import Thread


# Courtesy of https://stackoverflow.com/a/31614591/5615927
class PropagatingThread(Thread):
    """A ``Thread`` which propagates uncaught exceptions."""

    def run(self):
        self.exc = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self, timeout=None):
        super(PropagatingThread, self).join(timeout)
        if self.exc:
            raise self.exc
        return self.ret



from __future__ import annotations

from typing import Optional

from neoscore.core.layout_controllers import NewLine
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.point import ORIGIN, Point
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.text_alignment import AlignmentX
from neoscore.core.units import ZERO, Unit
from neoscore.western.abstract_staff import AbstractStaff
from neoscore.western.meter import Meter, MeterDef
from neoscore.western.staff import Staff
from neoscore.western.staff_object import StaffObject


class TimeSignature(PositionedObject, StaffObject):

    """A graphical time signature.

    Note that these time signatures are purely cosmetic; they have no effect on
    automatic engraving since this module has no internal concept of measures.
    """

    # Type sentinel used to hackily check type
    # without importing the type, risking cyclic imports.
    _neoscore_time_signature_type_marker = True

    def __init__(
        self,
        pos_x: Unit,
        staff: Staff,
        meter: MeterDef,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            pos_x: The x position relative to the parent staff
            staff: The parent staff
            meter: The meter represented.
            font: The font used. Defaults to the staff's font.
        """
        StaffObject.__init__(self, staff)
        PositionedObject.__init__(self, Point(pos_x, ZERO), staff)
        font = font or staff.music_font
        self._meter = Meter.from_def(meter)
        # Add one glyph for each digit
        self._upper_text = _TimeSignatureText(
            self.staff,
            self.x,
            ORIGIN,
            self,
            self.meter.upper_text_glyph_names,
            font=font,
            breakable=False,
        )
        self._lower_text = _TimeSignatureText(
            self.staff,
            self.x,
            ORIGIN,
            self,
            self.meter.lower_text_glyph_names,
            font=font,
            breakable=False,
        )
        self._position_glyphs()

    @property
    def upper_text(self) -> MusicText:
        """The upper glyph for the time signature"""
        return self._upper_text

    @property
    def lower_text(self) -> MusicText:
        """The lower glyph for the time signature"""
        return self._lower_text

    @property
    def visual_width(self) -> Unit:
        """The visual width of the time signature.

        This is useful for laying out staff objects near time signatures.
        """
        return self._visual_width

    @property
    def meter(self) -> Meter:
        """The meter represented.

        Setting this will automatically update the time signature's glyphs.
        """
        return self._meter

    @meter.setter
    def meter(self, value: MeterDef):
        self._meter = Meter.from_def(value)
        self.upper_text.text = self._meter.upper_text_glyph_names
        self.lower_text.text = self._meter.lower_text_glyph_names
        self._position_glyphs()

    def _position_glyphs(self):
        """This must be called after any modification to the glyph text"""
        # Vertically position, assuming time sig glyphs are 2 units tall and centered
        staff_center_y = self.staff.center_y
        if not self.meter.lower_text_glyph_names:
            self.upper_text.y = staff_center_y
        else:
            self.upper_text.y = staff_center_y - self.staff.unit(1)
            self.lower_text.y = staff_center_y + self.staff.unit(1)
        # Horizontally position
        upper_width = self.upper_text.bounding_rect.width
        lower_width = self.lower_text.bounding_rect.width
        if upper_width > lower_width:
            self._visual_width = upper_width
            self.lower_text.x += upper_width / 2
            self.lower_text.alignment_x = AlignmentX.CENTER
        elif lower_width > upper_width:
            self._visual_width = lower_width
            self.upper_text.x += lower_width / 2
            self.upper_text.alignment_x = AlignmentX.CENTER
        else:
            # Widths are equal. No adjustment needed.
            self._visual_width = upper_width


class _TimeSignatureText(MusicText):

    """A regular MusicText with special rendering behavior in staff fringes.

    If this text is rendered at the very beginning of a staff, it will shift according
    to the staff's fringe layout.
    """

    def __init__(
        self,
        staff: Optional[AbstractStaff],
        staff_pos_x: Optional[Unit],
        *args,
        **kwargs,
    ):
        self.staff = staff
        self.staff_pos_x = staff_pos_x
        super().__init__(*args, **kwargs)

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        if self.staff_pos_x == ZERO:
            fringe_layout = self.staff.fringe_layout_at(None)
            pos = Point(pos.x + fringe_layout.time_signature, pos.y)
        super().render_complete(pos, flowable_line, flowable_x)



from enum import Enum


class PenJoinStyle(Enum):

    """Styles controlling the shape of line joints on a pen stroke.

    The enum int values are for internal purposes and not guaranteed by the API.
    """

    # Values align with Qt's http://doc.qt.io/qt-5.9/qt.html#PenJoinStyle-enum

    BEVEL = 0x40
    """The triangular notch between the two lines is filled."""

    MITER = 0x00
    """The outer edges of the lines are extended to meet at an angle."""

    ROUND = 0x80
    """A circular arc between the two lines is filled."""



from __future__ import annotations

from typing import Optional, cast

from neoscore.core.brush import BrushDef
from neoscore.core.music_char import MusicChar
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicStringDef, MusicText
from neoscore.core.pen import PenDef
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.spanner_2d import Spanner2D
from neoscore.core.units import ZERO, Unit


class RepeatingMusicTextLine(Spanner2D, MusicText):

    """A spanner of repeating music text over its length.

    This automatically rotates the text to support 2D lines, but
    please note that rotated text breaking across flowable lines is
    not yet fully supported and will display incorrectly.
    """

    def __init__(
        self,
        start: PointDef,
        start_parent: Optional[PositionedObject],
        end_pos: PointDef,
        end_parent: Optional[PositionedObject],
        text: MusicStringDef,
        start_cap_text: Optional[MusicStringDef] = None,
        end_cap_text: Optional[MusicStringDef] = None,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        background_brush: Optional[BrushDef] = None,
    ):
        """
        Args:
            start: The starting point.
            start_parent: If no font is given, this or one of its ancestors
                must implement :obj:`.HasMusicFont`.
            end_pos: The end point.
            end_parent: The parent for the ending position.
                If ``None``, defaults to ``self``. (Note the lack of a default
                argument here: unlike with most other spanners, ``None`` must be given
                explicitly.)
            text: The text to be repeated over the spanner. Can be given as a SMuFL
                glyph name, or other shorthand forms. See :obj:`.MusicStringDef`.
            start_cap_text: A text specifier for the start of the text. Useful for
                things like "tr" beginnings to trill lines. This can be provided in the
                same form as ``text``.
            end_cap_text: A text specifier for the end of the text. Especially useful
                for line terminators like arrows at the end of arppeggio lines.
                This can be provided in the same form as ``text``.
            font: If provided, this overrides any font found in the ancestor chain.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
            background_brush: Optional brush used to paint the text's bounding rect
                behind it.
        """
        # Start the MusicText with a single repetition, then after
        # superclasses are set up figure out how many repetitions are
        # needed to cover ``self.length`` and update the text
        # accordingly.
        MusicText.__init__(
            self,
            start,
            start_parent,
            text,
            font,
            brush,
            pen,
            background_brush=background_brush,
        )
        Spanner2D.__init__(self, end_pos, end_parent or self)
        self.rotation = self.angle
        single_repetition_chars = self.music_chars
        main_char_width = self._approx_width(single_repetition_chars)

        if start_cap_text:
            # Again need to hackily set temporary text value to work out the width
            self.text = start_cap_text
            start_cap_chars = self.music_chars
            start_cap_width = self.font.bounding_rect_of(self.text).width
        else:
            start_cap_chars = []
            start_cap_width = ZERO
        if end_cap_text:
            # Same idea...
            self.text = end_cap_text
            end_cap_chars = self.music_chars
            end_cap_width = self.font.bounding_rect_of(self.text).width
        else:
            end_cap_chars = []
            end_cap_width = ZERO
        main_reps_needed = round(
            cast(
                float,
                (self.spanner_2d_length - end_cap_width - start_cap_width)
                / main_char_width,
            )
        )
        # Now set the final resolved text
        self.music_chars = (
            start_cap_chars
            + (single_repetition_chars * main_reps_needed)
            + end_cap_chars
        )

    def _approx_width(self, chars: List[MusicChar]) -> Unit:
        # Try to hackily account for ligatures and variable advance widths by averaging
        # out the width of main chars repeated many times.
        reps = 20
        return (
            self.font.bounding_rect_of("".join(c.codepoint for c in chars) * reps).width
            / reps
        )



"""All custom exceptions used throughout neoscore."""


from typing import Optional


class MusicFontMetadataNotFoundError(Exception):
    """Exception raised when metadata for a music font can't be found."""


class MusicFontGlyphNotFoundError(Exception):
    """Exception raised when a glyph cannot be found in a MusicFont"""

    def __init__(self, glyph_name: str, alternate_number: Optional[int]):
        if alternate_number is None:
            self.message = f"Cannot find glyph '{glyph_name}'"
        else:
            self.message = f"Cannot find glyph '{glyph_name} [alt {alternate_number}]'"
        super().__init__(self.message)


class NoClefError(Exception):
    """Exception raised when no clef is present in a Staff where needed"""


class NoAncestorStaffError(Exception):
    """Exception raised when a StaffObject does not have an ancestor Staff"""


class DynamicStringError(Exception):

    """Exception raised when a dynamic string cannot be parsed."""

    def __init__(self, string, character):
        self.message = (
            'Cannot parse dynamic string "{}" - character "{}" unknown.'
        ).format(string, character)
        super().__init__(self.message)


class NoFlagNeededError(Exception):
    """Exception raised when a Flag is created with a non-flaggable duration"""

    def __init__(self, duration):
        self.message = "Cannot create a flag for {}".format(duration)
        super().__init__(self.message)


class FontRegistrationError(Exception):
    """Exception raised when a font is loaded from disk unsuccessfully."""

    def __init__(self, font_file_path, detail=None):
        """
        Args:
            font_file_path (str): The path to the font file which could
                not be registered.
            detail (str): Optional error details.
        """
        self.message = "Could not register font from file '{}'.{}".format(
            font_file_path, " " + detail if detail else ""
        )
        super().__init__(self.message)


class UnknownFontFamilyError(Exception):
    """
    Exception raised when an invalid font name is passed to a FontInterface.
    """


class ColorBoundsError(Exception):
    """Exception raised when a color channel is set out of bounds"""

    def __init__(self, value: int):
        self.message = f"Invalid color channel value {value}"
        super().__init__(self.message)


class InvalidIntervalError(Exception):
    """An exception raised when an invalid interval specifier is used."""


class InvalidPitchDescriptionError(Exception):
    """An exception raised when an invalid pitch specifier is used."""


class InvalidImageFormatError(Exception):
    """Raised when an image format cannot be determined."""


class ImageExportError(Exception):
    """Raised when low level image export fails."""


class NoAncestorWithMusicFontError(Exception):
    """Raised when a lookup expecting to find an ancestor with a MusicFont fails."""


class ImageLoadingError(Exception):
    """Raised when an image cannot be loaded"""



from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional, Tuple

from neoscore.core.point import Point


class MouseButton(Enum):
    """Enum for mouse buttons.

    This currently only supports left, middle, and right buttons.

    The enum integer values are arbitrary and may change in the future.
    """

    LEFT = auto()
    MIDDLE = auto()
    RIGHT = auto()


class MouseEventType(Enum):
    """Enum for mouse event types.

    The enum integer values are arbitrary and may change in the future.
    """

    MOVE = auto()
    """The mouse was moved"""

    PRESS = auto()
    """A mouse button was pressed"""

    RELEASE = auto()
    """A mouse button was released"""

    DOUBLE_CLICK = auto()
    """A mouse button was double clicked"""


@dataclass(frozen=True)
class MouseEvent:
    """A mouse input event."""

    event_type: MouseEventType
    """The type of the event"""

    button: Optional[MouseButton]
    """The button pressed, if any.

    For mouse move events, this is the button pressed down while moving. For press and
    double click events, this is the button that caused the event. For release events,
    this is ``None``.
    
    Mouse events with multiple simultaneous button presses will somewhat arbitrarily
    return the first button pressed in order left, right, then middle.
    """

    window_pos: Tuple[int, int]
    """The mouse position on the window in ``x, y`` pixels."""

    document_pos: Point
    """The document-space position of the mouse."""



from enum import Enum

from neoscore.western.accidental_type import AccidentalType


class KeySignatureType(Enum):
    """Common key signatures.

    Key signatures are specified as dicts between the 7 pitch letter names
    and either an ``AccidentalType`` or ``None``. They are enumerated here in
    the form ``[pitch letter][f|s]_[major|minor]``. For example:

    * c major is ``KeySignatureType.c_major``
    * f-sharp minor is ``KeySignatureType.fs_minor``
    * d-flat major is ``KeySignature.df_major``

    This enum defines the entire circle of fifths - from 0 flats to 7,
    and 0 sharps to 7.
    """

    C_MAJOR = {
        "a": None,
        "b": None,
        "c": None,
        "d": None,
        "e": None,
        "f": None,
        "g": None,
    }

    A_MINOR = C_MAJOR.copy()

    DF_MAJOR = {
        "a": AccidentalType.FLAT,
        "b": AccidentalType.FLAT,
        "c": None,
        "d": AccidentalType.FLAT,
        "e": AccidentalType.FLAT,
        "f": None,
        "g": AccidentalType.FLAT,
    }

    BF_MINOR = DF_MAJOR.copy()

    CS_MAJOR = {
        "a": AccidentalType.SHARP,
        "b": AccidentalType.SHARP,
        "c": AccidentalType.SHARP,
        "d": AccidentalType.SHARP,
        "e": AccidentalType.SHARP,
        "f": AccidentalType.SHARP,
        "g": AccidentalType.SHARP,
    }

    D_MAJOR = {
        "a": None,
        "b": None,
        "c": AccidentalType.SHARP,
        "d": None,
        "e": None,
        "f": AccidentalType.SHARP,
        "g": None,
    }

    B_MINOR = D_MAJOR.copy()

    EF_MAJOR = {
        "a": AccidentalType.FLAT,
        "b": AccidentalType.FLAT,
        "c": None,
        "d": None,
        "e": AccidentalType.FLAT,
        "f": None,
        "g": None,
    }

    C_MINOR = EF_MAJOR.copy()

    E_MAJOR = {
        "a": None,
        "b": None,
        "c": AccidentalType.SHARP,
        "d": AccidentalType.SHARP,
        "e": None,
        "f": AccidentalType.SHARP,
        "g": AccidentalType.SHARP,
    }

    CS_MINOR = E_MAJOR.copy()

    F_MAJOR = {
        "a": None,
        "b": AccidentalType.FLAT,
        "c": None,
        "d": None,
        "e": None,
        "f": None,
        "g": None,
    }

    D_MINOR = F_MAJOR.copy()

    GF_MAJOR = {
        "a": AccidentalType.FLAT,
        "b": AccidentalType.FLAT,
        "c": AccidentalType.FLAT,
        "d": AccidentalType.FLAT,
        "e": AccidentalType.FLAT,
        "f": None,
        "g": AccidentalType.FLAT,
    }

    EF_MINOR = GF_MAJOR.copy()

    FS_MAJOR = {
        "a": AccidentalType.SHARP,
        "b": None,
        "c": AccidentalType.SHARP,
        "d": AccidentalType.SHARP,
        "e": AccidentalType.SHARP,
        "f": AccidentalType.SHARP,
        "g": AccidentalType.SHARP,
    }

    DS_MINOR = FS_MAJOR.copy()

    G_MAJOR = {
        "a": None,
        "b": None,
        "c": None,
        "d": None,
        "e": None,
        "f": AccidentalType.SHARP,
        "g": None,
    }

    E_MINOR = G_MAJOR.copy()

    AF_MAJOR = {
        "a": AccidentalType.FLAT,
        "b": AccidentalType.FLAT,
        "c": None,
        "d": AccidentalType.FLAT,
        "e": AccidentalType.FLAT,
        "f": None,
        "g": None,
    }

    F_MINOR = AF_MAJOR.copy()

    A_MAJOR = {
        "a": None,
        "b": None,
        "c": AccidentalType.SHARP,
        "d": None,
        "e": None,
        "f": AccidentalType.SHARP,
        "g": AccidentalType.SHARP,
    }

    FS_MINOR = A_MAJOR.copy()

    BF_MAJOR = {
        "a": None,
        "b": AccidentalType.FLAT,
        "c": None,
        "d": None,
        "e": AccidentalType.FLAT,
        "f": None,
        "g": None,
    }

    G_MINOR = BF_MAJOR.copy()

    B_MAJOR = {
        "a": AccidentalType.SHARP,
        "b": None,
        "c": AccidentalType.SHARP,
        "d": AccidentalType.SHARP,
        "e": None,
        "f": AccidentalType.SHARP,
        "g": AccidentalType.SHARP,
    }

    GS_MINOR = B_MAJOR.copy()

    CF_MAJOR = {
        "a": AccidentalType.FLAT,
        "b": AccidentalType.FLAT,
        "c": AccidentalType.FLAT,
        "d": AccidentalType.FLAT,
        "e": AccidentalType.FLAT,
        "f": AccidentalType.FLAT,
        "g": AccidentalType.FLAT,
    }



from __future__ import annotations

from typing import Dict, List, NamedTuple, Optional, Type, Union, cast

from typing_extensions import TypeAlias

from neoscore.core.brush import BrushDef
from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.music_char import MusicChar, MusicCharDef
from neoscore.core.music_font import MusicFont
from neoscore.core.pen import PenDef
from neoscore.core.point import ORIGIN, PointDef
from neoscore.core.positioned_object import PositionedObject, render_cached_property
from neoscore.core.rect import Rect
from neoscore.core.text import Text
from neoscore.core.text_alignment import AlignmentX, AlignmentY
from neoscore.core.units import Unit


class _CachedTextGeometryKey(NamedTuple):
    text: str  # The key is the plain unicode text, not rich MusicChar list
    font: MusicFont
    scale: float


class _CachedTextGeometry(NamedTuple):
    bounding_rect: Rect


_GEOMETRY_CACHE: Dict[_CachedTextGeometryKey, _CachedTextGeometry] = {}


MusicStringDef: TypeAlias = Union[MusicCharDef, List[MusicCharDef]]
"""Argument specifying SMuFL ``MusicText`` strings.

This supports several forms for different use-cases. The most commonly
used form is a simple SMuFL canonical glyph name.

* A canonical SMuFL glyph name. This may be an empty string to indicate 0-length text.
* A tuple of a glyph name and a SMuFL alternate number.
* A fully defined :obj:`.MusicChar`.
* A list of any of these, including an empty list.
"""


class MusicText(Text, HasMusicFont):
    """Text written in SMuFL compliant music fonts.

    For many use-cases, ``MusicText`` strings will consist of a single character, but
    longer strings are supported too.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        text: MusicStringDef,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        scale: float = 1,
        rotation: float = 0,
        background_brush: Optional[BrushDef] = None,
        breakable: bool = True,
        alignment_x: AlignmentX = AlignmentX.LEFT,
        alignment_y: AlignmentY = AlignmentY.BASELINE,
        transform_origin: PointDef = ORIGIN,
    ):
        """
        Args:
            pos: The position of the text.
            parent: The parent of the glyph. If no ``font`` is given,
                this or one of its ancestors must implement :obj:`.HasMusicFont`.
            text: The text to display. Can be given as a SMuFL glyph name
                or other shorthand forms. See ``MusicStringDef``.
            font: The music font to be used. If not specified, ``parent`` must
                implement :obj:`.HasMusicFont` or have an ancestor which does.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
            scale: A scaling factor to be applied in addition to the size of the music font.
            rotation: Angle in degrees. Note that breakable rotated text is
                not currently supported.
            background_brush: Optional brush used to paint the text's bounding rect
                behind it.
            breakable: Whether this object should break across lines in
                :obj:`.Flowable` containers.
            alignment_x: The text's horizontal alignment relative to ``pos``.
                Note that text which is not ``LEFT`` aligned does not currently display
                correctly when breaking across flowable lines.
            alignment_y: The text's vertical alignment relative to ``pos``.
            transform_origin: The origin point for rotation and scaling transforms
        """
        if font is None:
            font = HasMusicFont.find_music_font(parent)
        self._music_chars = MusicText._resolve_music_chars(font, text)
        resolved_str = MusicText._music_chars_to_str(self._music_chars)
        Text.__init__(
            self,
            pos,
            parent,
            resolved_str,
            font,
            brush,
            pen,
            scale,
            rotation,
            background_brush,
            breakable,
            alignment_x,
            alignment_y,
            transform_origin,
        )

    @property
    def music_chars(self) -> List[MusicChar]:
        """A list of the SMuFL characters in the string including metadata.

        If set, this will also update ``text``.
        """
        return self._music_chars

    @music_chars.setter
    def music_chars(self, value: List[MusicChar]):
        self._music_chars = value
        self._text = MusicText._music_chars_to_str(value)

    @property
    def text(self) -> str:
        """The raw unicode representation of the SMuFL text.

        If set, this will also update ``music_chars``.
        """
        return self._text

    @text.setter
    def text(self, value: MusicStringDef):
        self._music_chars = MusicText._resolve_music_chars(self.music_font, value)
        resolved_str = MusicText._music_chars_to_str(self._music_chars)
        self._text = resolved_str

    @property
    def music_font(self) -> MusicFont:
        """The SMuFL font used in this text.

        This is an expressive synonym for :obj:`font <.Text.font>`.
        """
        return cast(MusicFont, self._font)

    @music_font.setter
    def music_font(self, value: MusicFont):
        self._font = value

    @property
    def unit(self) -> Type[Unit]:
        """A unit type where ``unit(1)`` is a standard staff space in the font."""
        return self.music_font.unit

    @render_cached_property
    def _raw_scaled_bounding_rect(self) -> Rect:
        key = _CachedTextGeometryKey(self.text, self.music_font, self.scale)
        cached_result = _GEOMETRY_CACHE.get(key)
        if cached_result:
            return cached_result.bounding_rect
        bounding_rect = self.font.bounding_rect_of(self.text) * self.scale
        _GEOMETRY_CACHE[key] = _CachedTextGeometry(bounding_rect)
        return bounding_rect

    @staticmethod
    def _music_chars_to_str(music_chars: List[MusicChar]) -> str:
        return "".join(char.codepoint for char in music_chars)

    @staticmethod
    def _resolve_music_chars(font: MusicFont, text: MusicStringDef) -> List[MusicChar]:
        if isinstance(text, list):
            music_chars = []
            for text_char in text:
                char = MusicText._resolve_single_music_char(
                    font, cast(MusicCharDef, text_char)
                )
                if char is not None:
                    music_chars.append(char)
            return music_chars
        else:
            char = MusicText._resolve_single_music_char(font, text)
            if char is None:
                return []
            else:
                return [char]

    @staticmethod
    def _resolve_single_music_char(
        font: MusicFont, char: MusicCharDef
    ) -> Optional[MusicChar]:
        if isinstance(char, str):
            if char:
                return MusicChar(font, char)
            else:
                return None
        elif isinstance(char, tuple):
            return MusicChar(font, *char)
        elif isinstance(char, MusicChar):
            return char
        else:
            raise TypeError



import os
from time import time
from typing import Optional, Tuple

from PyQt5 import QtCore, QtWidgets, uic

QT_PRECISE_TIMER = 0


class MainWindow(QtWidgets.QMainWindow):
    """The primary entry point for all UI code.

    This bootstraps the ``main_window.ui`` structure.
    """

    _ui_path = os.path.join(os.path.dirname(__file__), "main_window.ui")

    def __init__(self):
        super().__init__()
        uic.loadUi(MainWindow._ui_path, self)
        self.refresh_func = None
        self.mouse_event_handler = None
        self._frame = 0  # Frame counter used in debug mode

    def show(
        self,
        min_size: Optional[Tuple[int, int]] = None,
        max_size: Optional[Tuple[int, int]] = None,
        fullscreen: bool = False,
    ):
        if min_size:
            self.setMinimumSize(QtCore.QSize(min_size[0], min_size[1]))
        if max_size:
            self.setMaximumSize(QtCore.QSize(max_size[0], max_size[1]))
        QtCore.QTimer.singleShot(0, QT_PRECISE_TIMER, self.refresh)  # noqa
        if fullscreen:
            super().showFullScreen()
        else:
            super().show()

    @QtCore.pyqtSlot()
    def refresh(self):
        start_time = time()
        if self.refresh_func:
            requested_delay_s = self.refresh_func(start_time)
            requested_delay_ms = int(requested_delay_s * 1000)
            QtCore.QTimer.singleShot(
                requested_delay_ms, QT_PRECISE_TIMER, self.refresh  # noqa
            )
            # if env.DEBUG:
            #     update_time = time() - start_time
            #     refresh_fps = int(1 / (time() - start_time))
            #     if self._frame % 30 == 0:
            #         print(
            #             f"Scene update took {int(update_time * 1000)} ms ({refresh_fps} / s)"
            #         )
            #         print(f"requested delay was {requested_delay_ms} ms")
            #     self._frame += 1
        # The viewport is unconditionally updated because we disable automatic updates
        self.graphicsView.viewport().update()



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional
from warnings import warn

from PyQt5.QtWidgets import QGraphicsItem

from neoscore.core import neoscore
from neoscore.core.point import Point


@dataclass(frozen=True)
class PositionedObjectInterface:
    """Interface for a generic graphic object.

    All graphic interfaces for renderable objects should descend from
    this and also be immutable dataclasses.
    """

    pos: Point
    """The position of the object.

    If a parent is provided, this position is relative to that interface. Otherwise it
    is in absolute document coordinates.
    """

    parent: Optional[PositionedObjectInterface]
    """The object's parent, if any.

    If a parent interface is provided, it must be rendered before this interface.
    """

    scale: float
    """A scaling factor, where 1 is no scaling.

    This occurs relative to ``transform_origin``.

    Scaling is inherited from parents to children along the interface tree.
    """

    rotation: float
    """Rotation angle in degrees, where 0 is no rotation.

    This occurs relative to ``transform_origin``.

    Rotation is inherited from parents to children along the interface tree.
    """

    transform_origin: Point
    """The origin point for rotation and scaling transforms"""

    _qt_object: Optional[QGraphicsItem] = field(init=False, compare=False, repr=False)
    """A corresponding Qt object for internal use only.

    This value is set during rendering and is not meant to be set more than once.
    """

    def render(self):
        """Render the object to the scene.

        This is typically done by constructing a `QGraphicsItem` subclass and calling
        `_register_qt_object` with it. Do *not* manually assign the Qt object's parent
        or add it to the Qt scene.
        """
        raise NotImplementedError

    def _parent_qt_obj(self) -> Optional[QGraphicsItem]:
        if self.parent:
            parent_qt_obj = getattr(self.parent, "_qt_object", None)
            if not parent_qt_obj:
                warn(
                    "Parent interface was provided but corresponding Qt object"
                    + f" not available when needed for {self}"
                )  # implicitly return None
            return parent_qt_obj
        return None

    def _register_qt_object(self, obj: QGraphicsItem):
        parent_obj = self._parent_qt_obj()
        if parent_obj:
            obj.setParentItem(parent_obj)
        else:
            neoscore.app_interface.scene.addItem(obj)
        super().__setattr__("_qt_object", obj)



from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.path import Path
from neoscore.core.pen import Pen
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO, Unit


class LedgerLine(Path, HasMusicFont):

    """A staff ledger line.

    These are generated automatically by :obj:`.Chordrest` objects,
    but can be manually instantiated as well.
    """

    def __init__(self, pos: PointDef, parent: PositionedObject, base_length: Unit):
        """
        Args:
            pos: A position at the left edge of the notehead column.
            parent: This or one of its ancestors must implement :obj:`.HasMusicFont`.
            base_length: The of the notehead this line is related to.
                The real length will be this plus a small extension defined in the
                :obj:`.MusicFont`'s engraving defaults.
        """
        Path.__init__(self, pos, parent=parent)
        font = self.music_font
        thickness = font.engraving_defaults["legerLineThickness"]
        self.pen = Pen(thickness=thickness)
        extension = font.engraving_defaults["legerLineExtension"]
        length = base_length + extension
        self.move_to(-extension, ZERO)
        self.line_to(length, ZERO)



"""Enums describing text alignment.

All enum int values are arbitrary and not part of the guaranteed API.
"""

from enum import Enum, auto


class AlignmentX(Enum):
    """Enum defining horizontal text alignments."""

    LEFT = auto()
    CENTER = auto()
    RIGHT = auto()


class AlignmentY(Enum):
    """Enum defining vertical text alignments.

    Only baseline and center alignments are currently supported.
    """

    BASELINE = auto()
    CENTER = auto()
